/* NickServ functions.
 *
 * Auspice Services is copyright (c) 2000-2001 In Mean
 *     E-mail: auspice@khmerpop.com
 *
 * SirvNET Services is copyright (c) 1998-2000 Trevor Klingbeil.
 *     E-mail: <priority1@dal.net>
 * Originally based on EsperNet Services(c) by Andy Church.
 * This program is free but copyrighted software; see the file LICENSE for
 * details.
 */

#include "../inc/services.h"
#include "../inc/timeout.h"

static NickInfo *nicklists[256];	/* One for each initial character */

#include "ns-help.c"


/* For local timeout use: */
#define TO_COLLIDE	0	/* Collide the user with this nick */
#define TO_RELEASE	1	/* Release a collided nick */
#define TO_COLLIDE_TL   2       /* Time left warning to collide */
#define TO_ENFORCE      3       /* Enforce a nick to guest* */

static void alpha_insert_nick(NickInfo *ni);
static NickInfo *makenick(const char *nick);
NickInfo * host(NickInfo * ni);
NickInfo * slave(const char *nick, int num);
int issibling (NickInfo * ni, const char *target);
int slavecount (const char *nick);
static void do_slaves(const char *source);
int userisnick (const char *nick);
long getflags (NickInfo * ni);
static void do_slavedrop (const char *source);

static int delnick(NickInfo *ni);
static void do_unhold(const char *source, const char *nick);
static void do_unmark(const char *source, const char *nick);

static void collide(NickInfo *ni, int from_timeout);
static void release(NickInfo *ni, int from_timeout);
static void add_ns_timeout(NickInfo *ni, int type, time_t delay);
static void del_ns_timeout(NickInfo *ni, int type);

static void update_last_seen_time(Timeout *to);
static void do_set_sex(NickInfo * ni, char *param);
static void do_set_malay(NickInfo *ni, char *param);
static void do_set_auth(NickInfo *ni, char *param);
static void do_set_privmsg(NickInfo *ni, char *param);
static void do_ohelp(const char *source);
static void do_help(const char *source);
static void do_register(const char *source);
static void do_identify(const char *source);
static void do_delete(const char *source);
static void do_drop(const char *source);
static void do_set(const char *source);
static void do_set_password(NickInfo *ni, char *param);
static void do_set_url(NickInfo *ni, char *param);
static void do_set_email(NickInfo *ni, char *param);
static void do_set_uin (NickInfo *ni, char *param);
static void do_set_age (NickInfo * ni, char *param);
static void do_set_info (NickInfo * ni, char *param);
static void do_set_kill(NickInfo *ni, char *param);
static void do_set_secure(NickInfo *ni, char *param);
static void do_set_private(NickInfo *ni, char *param);
static void do_set_emailmemos(NickInfo *ni, char *param);
static void do_set_neverop(NickInfo *ni, char *param);
static void do_set_noop(NickInfo *ni, char *param);
static void do_set_nomemo(NickInfo *ni, char *param);
static void do_set_nohost(NickInfo *ni, char *param);
static void do_access(const char *source);
static void do_link(const char *source);
static void do_info(const char *source);
static void do_acc(const char *source);
static void do_flist(const char *source);
static void do_list(const char *source);
static void do_recover(const char *source);
static void do_release(const char *source);
static void do_ghost(const char *source);
static void do_getpass(const char *source);
static void do_sendpass(const char *source);
static void do_forbid(const char *source);
static void do_unforbid(const char *source, const char *nick);
static char *get_os_access(NickInfo *ni);
static void do_setpass(const char *source);
static void do_set_hidemail(NickInfo *ni, char *param);
static void do_setflags(const char *source);
static void do_set_ircop(NickInfo *ni, char *param);
static void do_set_botmsg(NickInfo *ni, char *param);
static void do_setmlock(const char *source);
static void do_unfreeze(const char *source, const char *nick);

// static void do_set_mlock(NickInfo * ni, char *param);

/*************************************************************************/

/* Display total number of registered nicks and info about each; or, if
 * a specific nick is given, display information about that nick (like
 * /msg NickServ INFO <nick>).  If count_only != 0, then only display the
 * number of registered nicks (the nick parameter is ignored).
 */

void listnicks(int count_only, const char *nick)
{
    long count = 0;
    NickInfo *ni;
    int i;

    if (count_only) {

	for (i = 33; i < 256; ++i) {
	    for (ni = nicklists[i]; ni; ni = ni->next)
		++count;
	}
	printf("%d nicknames registered.\n", count);

    } else if (nick) {

	struct tm tm;
	char buf[512];

	if (!(ni = findnick(nick))) {
	    printf("%s not registered.\n");
	    return;
	} else if (ni->flags & NI_VERBOTEN) {
	    printf("%s is FORBIDden.\n");
	    return;
	}
	printf("%s is %s\n", nick, ni->last_realname);
	printf("Last seen address: %s\n", ni->last_usermask);
	tm = *localtime(&ni->time_registered);

	printf("  Time registered: %s %2d %02d:%02d:%02d %d\n",
			month_name(tm.tm_mon+1), tm.tm_mday, tm.tm_hour,
			tm.tm_min, tm.tm_sec, tm.tm_year);
	tm = *localtime(&ni->last_seen);
	printf("   Last seen time: %s %2d %02d:%02d:%02d %d\n",
			month_name(tm.tm_mon+1), tm.tm_mday, tm.tm_hour,
			tm.tm_min, tm.tm_sec, tm.tm_year);
	if (ni->url)
	    printf("              URL: %s\n", ni->url);
        if (ni->email && (ni->flags & NI_HIDE_EMAIL))
            printf("   E-mail address: Hidden\n");
        else
	    printf("   E-mail address: %s\n", ni->email);
	if (ni->uin)
	    printf("              UIN: %s\n", ni->uin);
	*buf = 0;
	if (ni->flags & NI_KILLPROTECT)
	    strcat(buf, "Kill protection");
	if (ni->flags & NI_SECURE) {
	    if (*buf)
		strcat(buf, ", ");
	    strcat(buf, "Security");
	}
	if (ni->flags & NI_PRIVATE) {
	    if (*buf)
		strcat(buf, ", ");
	    strcat(buf, "Private");
	}
        if (ni->flags & NI_MARK) {
            if (*buf)
                strcat(buf, ", ");
            strcat(buf, "Marked");
        }
        if (ni->flags & NI_HOLD) {
            if (*buf)
                strcat(buf, ", ");
            strcat(buf, "Held");
        }
	if (!*buf)
	    strcpy(buf, "None");
	printf("          Options: %s\n", buf);

    } else {

	for (i = 33; i < 256; ++i) {
	    for (ni = nicklists[i]; ni; ni = ni->next) {
		printf("    %-20s  %s\n", ni->nick, ni->last_usermask);
		++count;
	    }
	}
	printf("%d nicknames registered.\n", count);

    }
}
/*************************************************************************/
void listmemos(const char *nick, const char *pass)
{
    long count = 0;
    NickInfo *ni;
    int i;
    struct tm tm;
    char buf[512];
        if (!(ni = findnick(nick))) {
            printf("%s not registered.\n");
            return;
        } else if (ni->flags & NI_VERBOTEN) {
            printf("%s is FORBIDden.\n");
            return;
        }
	if (stricmp(ni->pass, pass) == 0) {
		do_webread(nick);
	} else {
		printf("not Match pass\n");
	}
}

/*************************************************************************/

NickInfo *retnick(int i)
{
    return nicklists[i];
}

/* Return information on memory use.  Assumes pointers are valid. */

void get_nickserv_stats(long *nrec, long *memuse)
{
    long count = 0, mem = 0;
    int i, j;
    NickInfo *ni;
    char **accptr;


    for (i = 0; i < 256; i++) {
	for (ni = nicklists[i]; ni; ni = ni->next) {
	    count++;
	    mem += sizeof(*ni);
	    if (ni->last_usermask)
		mem += strlen(ni->last_usermask)+1;
	    if (ni->last_realname)
		mem += strlen(ni->last_realname)+1;
            if (ni->url)
                mem += strlen(ni->url)+1;
            if (ni->email)
                mem += strlen(ni->email)+1;
            if (ni->uin)
                mem += strlen(ni->uin)+1;
            if (ni->forward)
                mem += strlen(ni->forward)+1;
            if (ni->mark)
                mem += strlen(ni->mark)+1;
            if (ni->hold)
                mem += strlen(ni->hold)+1;
            if (ni->forbid)
                mem += strlen(ni->forbid)+1;
            if (ni->age)
                mem += strlen(ni->age)+1;
            if (ni->info)
                mem += strlen(ni->info)+1;
            if (ni->sex)
                mem += strlen(ni->sex)+1;
            if (ni->mlock)
                mem += strlen(ni->mlock)+1;
/*
            if (ni->auth)
                mem += strlen(ni->auth)+1;
*/
	    mem += sizeof(char *) * ni->accesscount;
	    for (accptr=ni->access, j=0; j < ni->accesscount; accptr++, j++) {
		if (*accptr)
		    mem += strlen(*accptr)+1;
	    }
	}
    }
    *nrec = count;
    *memuse = mem;
}

/*************************************************************************/
/*************************************************************************/

/* Main NickServ routine. */

void nickserv(const char *source, char *buf)
{
    char *cmd, *s;
    NickInfo *ni = findnick(source);

    cmd = strtok(buf, " ");

    if (!cmd) {
	return;

    } else if (stricmp(cmd, "\1PING") == 0) {
	if (!(s = strtok(NULL, "")))
	    s = "\1";
	notice(s_NickServ, source, "\1PING %s", s);

    } else if (stricmp (cmd, "\1VERSION\1") == 0) {
    
        if (!(s = strtok (NULL, "")))
            s = "\1";
        notice(s_NickServ, source, "%sVERSION Auspice IRC Services %s", s, s);

    } else if (stricmp(cmd, "HELP") == 0) {
	do_help(source);

    } else if (stricmp(cmd, "REGISTER") == 0) {
	do_register(source);

    } else if (stricmp(cmd, "IDENTIFY") == 0) {
	do_identify(source);

    } else if (stricmp(cmd, "ID") == 0) {
        do_identify(source);

    } else if (stricmp(cmd, "LINK") == 0) {
	do_link(source);

    } else if (stricmp(cmd, "LINKS") == 0) {
	do_slaves(source);

    } else if (stricmp(cmd, "UNLINK") == 0) {
	do_slavedrop(source);

    } else if (stricmp(cmd, "DROP") == 0) {
	do_drop(source);

    } else if (stricmp(cmd, "SET") == 0) {
	do_set(source);

    } else if (stricmp(cmd, "GHOST") == 0) {
	do_ghost(source);

    } else if (stricmp(cmd, "ACCESS") == 0) {
	do_access(source);

    } else if (stricmp(cmd, "INFO") == 0) {
	do_info(source);

    } else if (stricmp(cmd, "ACC") == 0) {
        do_acc(source);

    } else if (stricmp(cmd, "RECOVER") == 0) {
	do_recover(source);

    } else if (stricmp(cmd, "RELEASE") == 0) {
	do_release(source);

    } else if (stricmp(cmd, "SELFDEL") == 0) {
	if (!ni) {
		notice(s_NickServ, source, NS_ONLYREG);	
	} else if (!(ni->flags & (NI_RECOGNIZED | NI_IDENTIFIED))) {
	        notice(s_NickServ, source, NS_PLZID, s_NickServ);
	        notice(s_NickServ, source, NS_IDCOM, s_NickServ);
	} else {
		do_selfdel(source);
	}

    } else if (stricmp(cmd, "LISTCHANS") == 0) {

	if (!ni) {
		notice(s_NickServ, source, NS_ONLYREG);	
	} else if (!(ni->flags & (NI_RECOGNIZED | NI_IDENTIFIED))) {
	        notice(s_NickServ, source, NS_PLZID, s_NickServ);
	        notice(s_NickServ, source, NS_IDCOM, s_NickServ);
	} else {
		if (ALLOWCLIST == 1) {
			do_glist(source);
	         } else {
			if (is_services_helpop(source)) {
				do_glist(source);
			} else {
				notice(s_NickServ, source, NS_DISABLED);
			} 
		}
	}

    } else if (stricmp(cmd, "WHOIS") == 0) {
	char *nick = strtok(NULL, " ");

	if (!nick) {
		notice(s_NickServ, source, "Syntax: /Msg NickServ WHOIS \2<Nick>\2");
        	return;
	} else {
           User *u = finduser(nick);
	   if(!u) {
		notice(s_NickServ, source, "%s: No Such nick",nick);
	   } else {
		NickInfo *ani;
		int temp_time;
		notice(s_NickServ, source, "[ %s ] (%s@%s) : %s",nick, u->username, u->host, u->realname);
		if (is_oper(source))
		notice(s_NickServ, source, "[ %s ] real host: %s",nick, u->realhost);
		ani = findnick(nick);
		if (ani) {
			if (ani->flags & NI_IDENTIFIED)
				notice(s_NickServ, source, "[ %s ] is Identified",nick);
			else
				notice(s_NickServ, source, "[ %s ] is not Identified yet",nick);
		} else {
		notice(s_NickServ, source, "[ %s ] is not a registered nick",nick);
		}
	        if (u->chans == NULL)
			notice(s_NickServ, source, "[ %s ] not join any channel",nick);
	        else {
		    struct u_chanlist *c, *c2;
	            char sbuf[512];
	            *sbuf = 0;
	            c = u->chans;
	            while (c) {
	                c2 = c->next;
	                if (*sbuf)
	                    strcat(sbuf, ", ");
	                strcat(sbuf, c->chan->name);
	                c = c2;
	             }
			notice(s_NickServ, source, "[ %s ] %s", nick, sbuf);
	
	        }

			notice(s_NickServ, source, "[ %s ] is on : %s",nick, u->server);
			if (is_netadmin(nick))
				notice(s_NickServ, source, "[ %s ] a Network Administrator",nick);
			else if (is_techadmin(nick))
				notice(s_NickServ, source, "[ %s ] a Technical Administrator",nick);
			else if (is_admin(nick))
				notice(s_NickServ, source, "[ %s ] a Server Administrator",nick);
			else if (is_oper(nick))
				notice(s_NickServ, source, "[ %s ] an IRC Operator",nick);
			if (is_oper(nick) && is_services_admin(nick))
				notice(s_NickServ, source, "[ %s ] a Services Administrator",nick);


		     temp_time = 0;
		     temp_time = time(NULL) - u->signon;

		     if (temp_time > 3600) {
		        temp_time /= 60;
			notice(s_NickServ, source, "[ %s ] Online Time: %d hour%s %d minute%s",nick,

		           temp_time/60, temp_time/60==1 ? "" : "s",
		           temp_time%60, temp_time%60==1 ? "" : "s");

		     } else {
			notice(s_NickServ, source, "[ %s ] Online Time: %d minute%s %d second%s",nick,
		        temp_time/60, temp_time/60==1 ? "" : "s",
		        temp_time%60, temp_time%60==1 ? "" : "s");
		     }


			notice(s_NickServ, source, "[ %s ] End of WHOIS list",nick);

  	    }
        } /* End of !nick */

    } else if (stricmp(cmd, "VERIFY") == 0) {
	char *nick = strtok(NULL, " ");

	if (!nick) {
		notice(s_NickServ, source, "Syntax: /Msg NickServ VERIFY \2<Nick>\2");
        	return;
	} else {
		NickInfo *ani;
		ani = findnick(nick);
		if (is_netadmin(nick))
			notice(s_NickServ, source, "[ %s ] is a Network Administrator",nick);
		else if (is_techadmin(nick))
			notice(s_NickServ, source, "[ %s ] is a Technical Administrator",nick);
		else if (is_admin(nick))
			notice(s_NickServ, source, "[ %s ] is a Server Administrator",nick);
		else if (is_oper(nick) && is_services_admin(nick))
			notice(s_NickServ, source, "[ %s ] a Services Administrator",nick);
		else if (is_oper(nick))
			notice(s_NickServ, source, "[ %s ] is an IRC Operator",nick);
		else if (ani && (ani->flags & (NI_HOP | NI_SOP)))
			notice(s_NickServ, source, "[ %s ] is an off-duty IRC Operator",nick);
		else
			notice(s_NickServ, source, "[ %s ] does not have any Operator Privelages",nick);
		
	} /* End of !nick */


    } else if (stricmp(cmd, "AUTH") == 0) {
	char *nick = strtok(NULL, " ");
	char *authcode = strtok(NULL, " ");

	if (!nick || !authcode) {
		notice(s_NickServ, source, "Syntax: /Msg NickServ AUTH  \2<Nick>\2 \2<Code>\2");
        	return;
	} else {
		NickInfo *ani;
		ani = findnick(nick);
		if (ani) {
/*
			if (stricmp(ni->auth,authcode)==0) {
				ani->flags &= ~NI_WAUTH;
			    notice(s_NickServ, source, "\2Authorize accept.\2");
			    notice(s_NickServ, source, NS_REGISTER, source, ani->access[0]);
		            notice(s_NickServ, source, NS_REGISTER2, ani->pass);
		            notice(s_NickServ, source, NS_REGMAIL, ani->email);
		            notice(s_NickServ, source, NS_KSECRET);
		            notice(s_NickServ, source, NS_HIDEMAILC, s_NickServ);
		            notice(s_NickServ, source, MISC_OFHELP, AOCHANS);
		            notice(s_NickServ, source, NS_INFOOWNNICK);
		            notice(s_NickServ, source, "http://%s", nsurl);
		            if (is_oper(source) && !(ani->flags & NI_HOP))
		                notice(s_NickServ, source, NS_TURNIRCOP);

			notice(s_NickServ, source, "Please type /Msg NickServ Identify %s", ani->pass);

			} else {
				notice(s_NickServ, source, "\2incorrect code\2");
	        		return;
			}
*/
		} else {
			notice(s_NickServ, source, "\2This nick is not a registered nick\2");
		}
	} /* End of !nick */


    } else if (!is_services_helpop(source)) {
	notice(s_NickServ, source, NS_UNKNOWC, cmd, s_NickServ);
        return;

    } else if (stricmp(cmd, "OHELP") == 0) {
	do_ohelp(source);

    } else if (stricmp(cmd, "SETFLAG") == 0) {
         do_setflags(source);

    } else if (!is_services_oper(source)) {
         notice(s_NickServ, source, PMD);

    } else if (stricmp(cmd, "FORBID") == 0) {
         do_forbid(source);

    } else if (!is_oper(source)) {
         notice(s_NickServ, source, PMD);
        return;

    } else if (stricmp(cmd, "SENDPASS") == 0) {
        do_sendpass(source);

    } else if (stricmp(cmd, "FLIST") == 0) {
        do_flist(source);

    } else if (!is_services_admin(source)) {
	notice(s_NickServ, source, NS_UNKNOWC, cmd, s_NickServ);

    } else if (stricmp(cmd, "GETPASS") == 0) {
	do_getpass(source);

    } else if (stricmp(cmd, "MARK") == 0) {
        char *nick = strtok(NULL, " ");
        NickInfo *ni;
        User *u = finduser(source);

        if ((!nick) || (stricmp(nick, "-") == 0)) {
            notice(s_NickServ,source,"Syntax: \2MARK\2 [-]<nick>");
            notice(s_NickServ,source,
		"\2/msg %s OHELP MARK\2 for more information",s_NickServ);
            return;
        }
        if (*nick == '-') {
           *nick++;
           do_unmark(source, nick);

	} else if (!(ni = findnick(nick))) {
            notice(s_NickServ, source, NS_NOTREG, nick);
        } else if (!(ni->flags & NI_MARK)) {
            ni->flags |= NI_MARK;
            if (ni->mark)
                  free(ni->mark);
            ni->mark = sstrdup(u->nick);
            slog("NS +M %s (%s!%s@%s)",
                nick,u->nick,u->username,u->host);
            do_break_log("NS_F", "NS +M %s (%s!%s@%s)",
                nick,u->nick,u->username,u->host);
            wallops(s_NickServ,
                "%s set the MARK flag for \2%s\2",u->nick,nick);
        } else {
            notice(s_NickServ,source, NS_ALREADYS, nick, "marked");
        }

    } else if (stricmp(cmd, "FREEZE") == 0) {
        char *nick = strtok(NULL, " ");
        NickInfo *ni;
        User *u = finduser(source);

        if ((!nick) || (stricmp(nick, "-") == 0)) {
            notice(s_NickServ,source,"Syntax: \2FREEZE\2 [-]<nick>");
            notice(s_NickServ,source,
		"\2/msg %s OHELP FREEZE\2 for more information",s_NickServ);
            return;
        }
        if (*nick == '-') {
           *nick++;
           do_unfreeze(source, nick);

	} else if (!(ni = findnick(nick))) {
            notice(s_NickServ, source, NS_NOTREG, nick);
        } else if (!(ni->flags & NI_FREEZE)) {
            ni->flags |= NI_FREEZE;
            slog("NS +FR %s (%s!%s@%s)",
                nick,u->nick,u->username,u->host);
            do_break_log("NS_F", "NS +FR %s (%s!%s@%s)",
                nick,u->nick,u->username,u->host);
            wallops(s_NickServ,
                "%s set the FREEZE flag for \2%s\2",u->nick,nick);
        } else {
            notice(s_NickServ,source, NS_ALREADYS, nick, "freeze");
        }

    } else if (stricmp(cmd, "HOLD") == 0) {
        char *nick = strtok(NULL, " ");
        NickInfo *ni;
        User *u = finduser(source);

        if ((!nick) || (stricmp(nick, "-") == 0)) {
            notice(s_NickServ,source,"Syntax: \2HOLD\2 [-]<nick>");
            notice(s_NickServ,source,
                ERR_OHELP,s_NickServ);
            return;
        }
        if (*nick == '-') {
           *nick++;
           do_unhold(source, nick);

        } else if (!(ni = findnick(nick))) {
            notice(s_NickServ, source, NS_NOTREG,nick);
        } else if (!(ni->flags & NI_HOLD)) {
            ni->flags |= NI_HOLD;
            if (ni->hold)
                  free(ni->hold);
            ni->hold = sstrdup(u->nick);
            slog("NS +H %s (%s!%s@%s)",
                nick,u->nick,u->username,u->host);
            do_break_log("NS_F", "NS +H %s (%s!%s@%s)",
                nick,u->nick,u->username,u->host);
            wallops(s_NickServ,
                "%s set the HELD flag for \2%s\2",u->nick,nick);
        } else {
            notice(s_NickServ,source,NS_ALREADYS,nick, "held");
        }

    } else if (!is_services_root(source)) {
            notice(s_NickServ,source,PMD);
            return;

    } else if (stricmp(cmd, "DELETE") == 0) {
        do_delete(source);

    } else if (stricmp(cmd, "SETPASS") == 0) {
        do_setpass(source);

    } else if (stricmp(cmd, "SETMLOCK") == 0) {
        do_setmlock(source);

    } else if (stricmp(cmd, "LIST") == 0) {
	do_list(source);

    } else {
        User *u = finduser(source);
	notice(s_NickServ, source,
		NS_UNKNOWC,
		cmd, s_NickServ);

        slog("NS *C %s (%s@%s) [%s]",
            source, u->username, u->host, cmd);
    }

}

/*************************************************************************/


/* Load/save data files. */

void load_ns_dbase(void)
{
    FILE *f;
    int ver, i, j, len;
    NickInfo *ni;

    if (!(f = open_db(s_NickServ, NICKSERV_DB, "r")))
	return;

    switch (ver = get_file_version(f, NICKSERV_DB)) {
      case 5:
        for (i = 33; i < 256; ++i) {
            while (fgetc(f) == 1) {
                   ni = smalloc(sizeof(NickInfo));
                   if (1 != fread(ni, sizeof(NickInfo), 1, f))
                       fatal_perror("Read error on %s", NICKSERV_DB);

                if (ni->flags & NI_ENCRYPTEDPW) {
                    /* Bail: it makes no sense to continue with encrypted
                    * passwords, since we won't be able to verify them */
                    fatal("%s: load database: password for %s encrypted "
                          "but encryption disabled, aborting",
                          s_NickServ, ni->nick);
                }

                ni->flags &= ~(NI_IDENTIFIED | NI_RECOGNIZED);
                alpha_insert_nick(ni);
/*
		if (DATAUP == 5) {
		    ni->auth = NULL;
		}
*/
		if (DATAUP == 3) {
		    ni->age = NULL;
		    ni->info = NULL;
		    ni->sex = NULL;
		    ni->mlock = NULL;
		    ni->sex = NULL;
//		    ni->auth = NULL;
		}
                if (ni->url)
                    ni->url = read_string(f, NICKSERV_DB);
                if (ni->email)
                    ni->email = read_string(f, NICKSERV_DB);
                if (ni->uin)
                    ni->uin = read_string(f, NICKSERV_DB);
                if (ni->age)
                    ni->age = read_string(f, NICKSERV_DB);
		if (ni->forward)
		    ni->forward = read_string(f, NICKSERV_DB);
                if (ni->hold)
                    ni->hold = read_string(f, NICKSERV_DB);
                if (ni->mark)
                    ni->mark = read_string(f, NICKSERV_DB);
                if (ni->forbid)
                    ni->forbid = read_string(f, NICKSERV_DB);
                if (ni->info)
                    ni->info = read_string(f, NICKSERV_DB);
                if (ni->mlock)
                    ni->mlock = read_string(f, NICKSERV_DB);
                if (ni->sex)
                    ni->sex = read_string(f, NICKSERV_DB);
/*
                if (ni->auth)
                    ni->auth = read_string(f, NICKSERV_DB);
*/
                ni->last_usermask = read_string(f, NICKSERV_DB);
                ni->last_realname = read_string(f, NICKSERV_DB);
                if (ni->accesscount) {
                    char **access;
                    access = smalloc(sizeof(char *) * ni->accesscount);
                    ni->access = access;
                    for (j = 0; j < ni->accesscount; j++, access++)
                        *access = read_string(f, NICKSERV_DB);
                }

                ni->id_timestamp = 0;
            } /* while (fgetc(f) == 1) */
        } /* for (i) */
        break;
      case 4:
      case 3:
      case 2:
      case 1:
      default:
	fatal("Unsupported version number (%d) on %s", i, NICKSERV_DB);
    } /* switch (version) */

    close_db(f, NICKSERV_DB);
}

void save_ns_dbase(void)
{
    FILE *f;
    int i, j, len;
    NickInfo *ni;
    char **access;

    if (!(f = open_db(s_NickServ, NICKSERV_DB, "w")))
	return;
    for (i = 33; i < 256; ++i) {
	for (ni = nicklists[i]; ni; ni = ni->next) {
	    fputc(1, f);
	    if (1 != fwrite(ni, sizeof(NickInfo), 1, f))
		fatal_perror("Write error on %s", NICKSERV_DB);
	    if (ni->url)
		write_string(ni->url, f, NICKSERV_DB);
	    if (ni->email)
		write_string(ni->email, f, NICKSERV_DB);
            if (ni->uin)
                write_string(ni->uin, f, NICKSERV_DB);
            if (ni->age)
                write_string(ni->age, f, NICKSERV_DB); 
	    if (ni->forward)
		write_string(ni->forward, f, NICKSERV_DB);
            if (ni->hold)
                write_string(ni->hold, f, NICKSERV_DB);
            if (ni->mark)
                write_string(ni->mark, f, NICKSERV_DB);
            if (ni->forbid)
                write_string(ni->forbid, f, NICKSERV_DB);
            if (ni->info)
                write_string(ni->info, f, NICKSERV_DB);
            if (ni->mlock)
                write_string(ni->mlock, f, NICKSERV_DB);
            if (ni->sex)
                write_string(ni->sex, f, NICKSERV_DB);
/*
            if (ni->auth)
                write_string(ni->auth, f, NICKSERV_DB);
*/
            if (ni->last_usermask)
	    write_string(ni->last_usermask ? ni->last_usermask : "",
							f, NICKSERV_DB);
            if (ni->last_realname)
	    write_string(ni->last_realname ? ni->last_realname : "",
							f, NICKSERV_DB);
	    for (access = ni->access, j = 0; j < ni->accesscount; ++access, ++j)
		write_string(*access, f, NICKSERV_DB);
	}
	fputc(0, f);
    }
    close_db(f, NICKSERV_DB);
}

/*************************************************************************/

/*************************************************************************/
/* Install any timeouts we need at startup. */

void ns_install_timeouts(void)
{
#if NS_LAST_SEEN_DELAY > 0
    add_timeout(NS_LAST_SEEN_DELAY, update_last_seen_time, 1);
#endif
}

/*************************************************************************/

/* Check whether a user is on the access list of the nick they're using.
 * If not, send warnings as appropriate.  If so (and not NI_SECURE), update
 * last seen info.  Return 1 if the user is valid and recognized, 0
 * otherwise (note that this means an NI_SECURE nick will always return 0
 * from here). */

int validate_user(User *u)
{
    NickInfo *ni, *hni;
    int on_access, i;

    if (!(ni = findnick (u->nick)) || !(hni = host (ni))) {
        send_cmd(s_NickServ, "SVSMODE %s -r 0", u->nick);
	return 0;
    }

    if (ni->flags & NI_WAUTH)
	return 0;

    if (ni->flags & NI_FREEZE) {
        send_cmd(s_NickServ, "SVSMODE %s -r 0", u->nick);
	return 0;
    }

    if (ni->flags & (NI_RECOGNIZED | NI_IDENTIFIED))
    {
        if (hni->last_usermask)
	        free (hni->last_usermask);
        hni->last_usermask = smalloc (strlen (u->nick) + strlen (u->username) + strlen (u->host) + 3);
        snprintf (hni->last_usermask, strlen (u->nick) + strlen (u->username) + strlen (u->host) + 3,
                                "%s@%s", u->username, u->host);

        return 0;
    }

    if (ni->flags & NI_VERBOTEN) {
        send_cmd(s_NickServ, "SVSMODE %s -r 0", u->nick);
	notice(s_NickServ, u->nick, NS_NEEDID);
	notice(s_NickServ, u->nick, NS_NEEDID2);
        add_ns_timeout(ni, TO_COLLIDE_TL, 30);
	return 0;
    }

    if (ni->id_timestamp != 0 && u->signon == ni->id_timestamp) {
	ni->flags |= NI_RECOGNIZED;
        ni->flags |= NI_IDENTIFIED;
        ni->last_seen = hni->last_seen = time(NULL);
	if (hni->last_usermask)
	    free(hni->last_usermask);
	hni->last_usermask = smalloc(strlen(u->username)+strlen(u->host)+2);
	sprintf(hni->last_usermask, "%s@%s", u->username, u->host);
	if (ni->last_realname)
 	    free(ni->last_realname);
	ni->last_realname = sstrdup(u->realname);
        send_cmd(s_NickServ, "SVSMODE %s +r 0", u->nick);
	if (!is_on_id_list(u->nick, hni->nick)) {
	    for (i=0; u->id_nicks[i] && i < MAX_IDS; i++)
		;
	    if (i < MAX_IDS)
	         u->id_nicks[i] = sstrdup(hni->nick);
	}
	return 1;
    }

    on_access = is_on_access(u, ni);

    if (!(hni->flags & NI_SECURE) &&
		(on_access || is_on_id_list(u->nick, u->nick) || is_on_id_list(u->nick, ni->last_usermask))) {

	ni->flags |= NI_RECOGNIZED;
	if (is_on_id_list(u->nick, u->nick) || is_on_id_list(u->nick, ni->last_usermask)) {
	    ni->flags |= NI_IDENTIFIED;
    	    ni->last_seen = hni->last_seen = time(NULL);
	    
	    if (hni->last_usermask)
	        free(hni->last_usermask);
	    hni->last_usermask = smalloc(strlen(u->username)+strlen(u->host)+2);
	    sprintf(hni->last_usermask, "%s@%s", u->username, u->host);
	    if (ni->last_realname)
 	        free(ni->last_realname);
	    ni->last_realname = sstrdup(u->realname);
        }
        send_cmd(s_NickServ, "SVSMODE %s +r 0", u->nick);
	return 1;
    }

    if (hni->flags & NI_SECURE) {
        send_cmd(s_NickServ, "SVSMODE %s -r 0", u->nick);
	notice(s_NickServ, u->nick, NS_NEEDID3);
	notice(s_NickServ, u->nick, NS_NEEDID4, s_NickServ);
	notice(s_NickServ, u->nick, NS_NEEDID5);
    } else {
        send_cmd(s_NickServ, "SVSMODE %s -r 0", u->nick);
	notice(s_NickServ, u->nick, NS_NEEDID6);
	notice(s_NickServ, u->nick, NS_NEEDID7, s_NickServ);
    }

    if ((hni->flags & NI_KILLPROTECT)
		&& !((hni->flags & NI_SECURE) && (on_access || is_on_id_list(u->nick, u->nick)))) {
	notice(s_NickServ, u->nick, NS_60SEC);
        add_ns_timeout(ni, TO_COLLIDE_TL, 30);
    }

    return 0;
}
/*************************************************************************/
static void do_slavedrop (const char *source)
{
    char *nick = strtok (NULL, " ");
    NickInfo *ni, *hni;
    User *u = finduser (source);

    if (!nick)
    {
        notice (s_NickServ, source, "Syntax: \2UNLINK Nick\2");
        notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "UNLINK");
    }
    else if (stricmp (source, nick) == 0)
        notice (s_NickServ, source, ERR_NOT_ON_YOURSELF, "UNLINK");
    else if (!(ni = findnick (nick)) || !(hni = host (ni)))
        notice (s_NickServ, source, NS_NOT_REGISTERED, nick);

    else if (stricmp (hni->nick, source) != 0)
    {
        if (issibling(findnick(source), nick))
            notice (s_NickServ, source, ERR_MUST_BE_HOST, "UNLINK");
        else
            notice (s_NickServ, source, ERR_ACCESS_DENIED);
    }
    else if (!u || !(hni->flags & NI_IDENTIFIED))
    {
        notice (s_NickServ, source, ERR_NEED_PASSWORD, "UNLINK");
        notice (s_NickServ, source, ERR_IDENT_NICK_FIRST, s_NickServ);
        if (!u)
            log ("[ERR] %s: UNLINK from nonexistent user %s", s_NickServ, source);
    }
    else
    {
        delnick (ni);
        log ("[DROP] %s: %s!%s@%s dropped nickname %s (slave)", s_NickServ,
                   source, u->username, u->host, nick);
        notice (s_NickServ, source, NS_DROPPED, nick, "", "", "");
    }
}

/*************************************************************************/

/* Cancel validation flags for a nick (i.e. when the user with that nick
 * signs off or changes nicks).  Also cancels any impending collide. */

void cancel_user(User *u)
{
    NickInfo *ni;

    if (ni = findnick(u->nick)) {
	ni->flags &= ~(NI_IDENTIFIED | NI_RECOGNIZED);
	del_ns_timeout(ni, TO_COLLIDE);
    }
}

/*************************************************************************/

/* Remove all nicks which have expired. */

void expire_nicks()
{
    NickInfo *ni, *hni, *next;
    int i, j;
    const time_t expire_time = NICK_EXPIRE*24*60*60;
    const time_t expire_mail = NICK_EXPIRE_MAIL*24*60*60;
    time_t now = time(NULL);
    long count = 0;
    long xcount = 0;
    FILE *fp;
    char cmdbuf[500];
    char cmdbuf2[500];
    char timebuf[64];
    struct tm tm;

    if (allowmail == 1) {
	    tm = *localtime(&now);
	    strftime(timebuf, sizeof(timebuf), "%a %b %d %H:%M:%S %Y %Z", &tm);
	    timebuf[sizeof(timebuf)-1] = 0;
    }

    for (i = 33; i < 256; ++i) {
	for (ni = nicklists[i]; ni; ni = next) {
	    next = ni->next;
            hni = host(ni);
	    if (ni->last_seen)
	    if ((now - ni->last_seen >= (expire_time - expire_mail))
					&& !slavecount (ni->nick)
			                && !(hni->flags & (NI_VERBOTEN | NI_HOLD | NI_HOP | NI_EXMAIL))) {

                slog("NS XE %s", ni->nick);
                do_break_log("NS_X", "NS XE %s", ni->nick);
		log("E-mail Near Expiring nickname %s", ni->nick);
	        ni->flags |= NI_EXMAIL;

      if (allowmail == 1) {
	      fp = fopen("memo.txt", "a");
	        fprintf (fp, "Date: %s\n", timebuf);
	        fprintf (fp, "From: %s <%s>\n", s_NickServ, PASS_EMAIL);
	        fprintf (fp, "To: %s <%s>\n", ni->nick, ni->email);
	        fprintf (fp, "Subject: %s Nickname Nearly Expire Reminder!!\n", s_NickServ);
	        fprintf (fp, "You received this email because you have registered\n");
	        fprintf (fp, "Your nickname (%s) on the %s\n\n",ni->nick, NETWORK_NAME);
	        fprintf (fp, "with password: %s\n\n", ni->pass);        
		fprintf (fp, "Your nick will expire in %d days if you don't identify.\n\n", NICK_EXPIRE_MAIL);
	        fprintf (fp, "THANKS YOU FOR USING THE %s IRC NETWORK\n\n", NETWORK_NAME);
	        fprintf (fp, "Sincerely,\n");
	        fprintf (fp, "%s (%s)\n\n\n", ADMIN_NAME, SERVICES_MASTER);
	        fprintf (fp, "----------------------------\n");
//	        fprintf (fp, "Get Free Web Base EMail from http://mail.pleaselook.com\n");
	      fclose (fp);

	      sprintf (cmdbuf, "%s -t %s < %s", sendmail_path, ni->email, "memo.txt");
	      sprintf (cmdbuf2, "rm -f %s", "memo.txt");

	      system(cmdbuf);
	      system(cmdbuf2);
	} /* If allow mail */
    } /* for ni */

	    if (ni->last_seen)
	    if ((now - ni->last_seen >= expire_time
					&& !slavecount (ni->nick)
			                && !(host(ni)->flags & (NI_VERBOTEN | NI_HOLD | NI_HOP))) || ((now - ni->last_seen > 1) && (ni->flags & NI_WAUTH))) {
		++xcount;
                slog("NS X %s", ni->nick);
                do_break_log("NS_X", "NS X %s", ni->nick);
		do_expireanick(ni->nick);
		log("Expiring nickname %s", ni->nick);
		del_hostline(ni->nick);
		delnick(ni);
	    }
	}

    } /* Main For */

    for (i = 33; i < 256; ++i) {
        for (ni = nicklists[i]; ni; ni = ni->next)
            ++count;
    }

    if (DISPLAY_UPDATES == 1)
        wallops(SERVER_NAME, "Completed NickName Expire (%d/%d)",
        	xcount,count);
}


/*************************************************************************/

/* Return the NickInfo structure for the given nick, or NULL if the nick
 * isn't registered. */

NickInfo *findnick(const char *nick)
{
    NickInfo *ni;

    for (ni = nicklists[tolower(*nick)]; ni; ni = ni->next) {
	if (stricmp(ni->nick, nick) == 0)
	    return ni;
    }
    return NULL;
}

/*************************************************************************/
/*********************** NickServ private routines ***********************/
/*************************************************************************/

/* Is the given user's address on the given nick's access list?  Return 1
 * if so, 0 if not. */

int is_on_access(User *u, NickInfo *ni)
{
    int i;
    char *buf;
    NickInfo *hni = host (ni);

    if (!ni || !hni) return 0;
    i = strlen(u->username);
    buf = smalloc(i + strlen(u->host) + 2);
    sprintf(buf, "%s@%s", u->username, u->host);
    strlower(buf+i+1);

    for (i = 0; i < hni->accesscount; ++i) {
	if (match_wild(hni->access[i], buf)) {
	    free(buf);
	    return 1;
	}
    }
    free(buf);
    return 0;
}

/*************************************************************************/

/* Insert a nick alphabetically into the database. */

static void alpha_insert_nick(NickInfo *ni)
{
    NickInfo *ni2, *ni3;
    char *nick = ni->nick;

    for (ni3 = NULL, ni2 = nicklists[tolower(*nick)];
			ni2 && stricmp(ni2->nick, nick) < 0;
			ni3 = ni2, ni2 = ni2->next)
	;
    ni->prev = ni3;
    ni->next = ni2;
    if (!ni3)
	nicklists[tolower(*nick)] = ni;
    else
	ni3->next = ni;
    if (ni2)
	ni2->prev = ni;
}

/*************************************************************************/

/* Add a nick to the database.  Returns a pointer to the new NickInfo
 * structure if the nick was successfully registered, NULL otherwise.
 * Assumes nick does not already exist. */

static NickInfo *makenick(const char *nick)
{
    NickInfo *ni;
    User *u;

    ni = scalloc(sizeof(NickInfo), 1);
    strscpy(ni->nick, nick, NICKMAX);
    alpha_insert_nick(ni);
    return ni;
}

/*************************************************************************/

/* Remove a nick from the NickServ database.  Return 1 on success, 0
 * otherwise.  Also deletes the nick from any channel access lists it is
 * on. */

static int delnick(NickInfo *ni)
{
    User *u;
    int i, cnt = 1;

    for (i = slavecount (ni->nick); i; --i)
        cnt += delnick (slave (ni->nick, 1));

    for (u = userlist; u; u = u->next) {
       for (i=0; u->id_nicks[i] && i < MAX_IDS; i++) {
          if (stricmp(u->id_nicks[i], ni->nick) == 0)
              free(u->id_nicks[i]);
       }
    }

    if (ni->next)
	ni->next->prev = ni->prev;
    if (ni->prev)
	ni->prev->next = ni->next;
    else
	nicklists[tolower(*ni->nick)] = ni->next;
    if (ni->last_usermask)
	free(ni->last_usermask);
    if (ni->last_realname)
	free(ni->last_realname);
    if (ni->access) {
	for (i = 0; i < ni->accesscount; ++i)
	    free(ni->access[i]);
	free(ni->access);
    }

    if (ni->url)
      free(ni->url);
    if (ni->email)
      free(ni->email);
    if (ni->forward)
      free(ni->forward);
    if (ni->hold)
      free(ni->hold);
    if (ni->mark);
      free(ni->mark);
    if (ni->forbid)
      free(ni->forbid);
    if (ni->uin)
      free(ni->uin);
    if (ni->age)
      free(ni->age);
    if (ni->info)
      free(ni->info);
    if (ni->sex)
      free(ni->sex);
    if (ni->mlock)
      free(ni->mlock);
/*
    if (ni->auth)
      free(ni->auth);
*/
    cs_remove_nick(ni->nick);
    clear_memos(ni->nick);
    os_remove_nick(ni->nick);
    free(ni);
//    expire_ur_chans();
    return 1;
}

/*************************************************************************/

/* Collide a nick. */

static void collide(NickInfo *ni, int from_timeout)
{
    User *u;
    char newnick[NICKMAX];
    long int guestbuf = 0;
    double x = RAND_MAX + 1.0;
    int y;

    y = 1 + rand() * (99999 / x);
    guestbuf = y;
    snprintf (newnick, sizeof (newnick), "Guest%d", guestbuf);

    if (!from_timeout)
	del_ns_timeout(ni, TO_COLLIDE);

    add_ns_timeout(ni, TO_ENFORCE, 1);
    
    send_cmd(NULL, "SVSNICK %s %s :%lu", ni->nick, newnick, time(NULL));
    notice(s_NickServ, newnick, "For more information on what this means, go to:");
    notice(s_NickServ, newnick, "http://%s", guesturl);

}
/*************************************************************************/
static inline void enforce(NickInfo *ni, int from_timeout)
{
   if (!from_timeout)
       del_ns_timeout(ni, TO_ENFORCE);

   send_cmd(NULL, "NICK %s 1 %lu enforcer %s %s 0 :Nick Protection Enforcement",
   ni->nick, time(NULL), services_host, server_name);

   ni->flags |= NI_KILL_HELD;
   add_ns_timeout(ni, TO_RELEASE, RELEASE_TIMEOUT);
}
/*************************************************************************/

/*************************************************************************/

/* Release hold on a nick. */

static void release(NickInfo *ni, int from_timeout)
{
    if (!from_timeout)
	del_ns_timeout(ni, TO_RELEASE);
    send_cmd(ni->nick, "QUIT :Leaving");
    ni->flags &= ~NI_KILL_HELD;
}

/*************************************************************************/
/*************************************************************************/

static struct my_timeout {
    struct my_timeout *next, *prev;
    NickInfo *ni;
    Timeout *to;
    int type;
} *my_timeouts;

/*************************************************************************/
/* Remove a collide/release timeout from our private list. */

static void rem_ns_timeout(NickInfo *ni, int type)
{
    struct my_timeout *t, *t2;

    t = my_timeouts;
    while (t) {
	if (t->ni == ni && t->type == type) {
	    t2 = t->next;
	    if (t->next)
		t->next->prev = t->prev;
	    if (t->prev)
		t->prev->next = t->next;
	    else
		my_timeouts = t->next;
	    free(t);
	    t = t2;
	} else {
	    t = t->next;
	}
    }
}
/*************************************************************************/
static void timeout_enforce(Timeout *t)
{
    NickInfo *ni = t->data;

    rem_ns_timeout(ni, TO_ENFORCE);
    enforce(ni, 1);
}
/*************************************************************************/

/* Collide a nick on timeout. */

static void timeout_collide(Timeout *t)
{
    NickInfo *ni = t->data;
    User *u;

    rem_ns_timeout(ni, TO_COLLIDE);
    /* If they identified or don't exist anymore, don't kill them. */
    if (ni->flags & NI_IDENTIFIED
		|| !(u = finduser(ni->nick))
		|| u->my_signon > t->settime)
	return;
    /* The RELEASE timeout will always add to the beginning of the
     * list, so we won't see it.  Which is fine because it can't be
     * triggered yet anyway. */
    collide(ni, 1);
}

static void timeout_collide_tl(Timeout *t)
{
   NickInfo *ni = t->data;
   User *u;

   rem_ns_timeout(ni, TO_COLLIDE_TL);
   if (ni->flags & NI_IDENTIFIED
		|| !(u = finduser(ni->nick))
		|| u->my_signon > t->settime)
	return;
   add_ns_timeout(ni, TO_COLLIDE, 30);
   notice(s_NickServ, ni->nick, NS_30SEC);
}

/*************************************************************************/

/* Release a nick on timeout. */

static void timeout_release(Timeout *t)
{
    NickInfo *ni = t->data;

    rem_ns_timeout(ni, TO_RELEASE);
    release(ni, 1);
}

/*************************************************************************/
/* Add a collide/release timeout. */

static void add_ns_timeout(NickInfo *ni, int type, time_t delay)
{
    Timeout *to;

    struct my_timeout *t;
    void (*timeout_routine)(Timeout *);

    if (type == TO_COLLIDE_TL)
        timeout_routine = timeout_collide_tl;
    else if (type == TO_COLLIDE)
	timeout_routine = timeout_collide;
    else if (type == TO_RELEASE)
	timeout_routine = timeout_release;
    else if (type == TO_ENFORCE)
        timeout_routine = timeout_enforce;
    else {
	log("NickServ: unknown timeout type %d!  ni=%p (%s), delay=%ld",
		type, ni, ni->nick, delay);
	return;
    }

    to = add_timeout(delay, timeout_routine, 0);
    to->data = ni;
    t = smalloc(sizeof(*t));
    t->next = my_timeouts;
    my_timeouts = t;
    t->prev = NULL;
    t->ni = ni;
    t->to = to;
    t->type = type;
}

/*************************************************************************/

/* Delete a collide/release timeout. */

static void del_ns_timeout(NickInfo *ni, int type)
{

    struct my_timeout *t, *t2;

    t = my_timeouts;
    while (t) {
	if (t->ni == ni && t->type == type) {
	    t2 = t->next;
	    if (t->next)
		t->next->prev = t->prev;
	    if (t->prev)
		t->prev->next = t->next;
	    else
		my_timeouts = t->next;
	    del_timeout(t->to);
	    free(t);
	    t = t2;
	} else {
	    t = t->next;
	}
    }  
}

/*************************************************************************/
/*************************************************************************/

/* Update last-seen time for all users currently online. */

static void update_last_seen_time(Timeout *to_unused)
{
    User *u;
    NickInfo *ni;

    /* This is O(n^2) at the moment, because the user list isn't known to
     * be sorted.
     */
    for (u = first_user(); u; u = u->next) {
	if (ni = findnick(u->nick))
            if (ni && (ni->flags & (NI_IDENTIFIED | NI_RECOGNIZED)))
        	 ni->last_seen = time(NULL);
    }
}

/*************************************************************************/
/*********************** NickServ command routines ***********************/
/*************************************************************************/

/* Return a help message. */

static void do_ohelp(const char *source)
{
    char *cmd = strtok(NULL, "");
    char buf[BUFSIZE];
    int is_servadmin = is_services_oper(source);

    if (!cmd)
        slog("NS - %s requested ohelp: (index)", source);

    if (cmd) {
        slog("NS - %s requested ohelp: %s", source, cmd);

        if (stricmp(cmd, "DELETE") == 0) {
            notice_list(s_NickServ, source, oper_delete_help);
            return;
        } else if (stricmp(cmd, "LIST") == 0) {
            notice_list(s_NickServ, source, oper_list_help);
            return;
        } else if (stricmp(cmd, "FLIST") == 0) {
            notice_list(s_NickServ, source, flist_help);
            return;
        } else if (stricmp(cmd, "GETPASS") == 0) {
            notice_list(s_NickServ, source, getpass_help);
            return;
        } else if (stricmp(cmd, "FORBID") == 0) {
            notice_list(s_NickServ, source, forbid_help);
            return;
        } else if (stricmp(cmd, "SENDPASS") == 0) {
            notice_list(s_NickServ, source, sendpass_help);
            return;
        } else if (stricmp(cmd, "MARK") == 0) {
            notice_list(s_NickServ, source, mark_help);
            return;
        } else if (stricmp(cmd, "HOLD") == 0) {
            notice_list(s_NickServ, source, hold_help);
            return;
        } else if (stricmp(cmd, "SETPASS") == 0) {
            notice_list(s_NickServ, source, setpass_help);
            return;
        } else if (stricmp(cmd, "SETMLOCK") == 0) {
            notice_list(s_NickServ, source, setmlock_help);
            return;
        } else if (stricmp(cmd, "SETFLAG") == 0) {
            notice_list(s_NickServ, source, setflags_help);
            return;
        }

    }

    if (!cmd) {
        notice_list(s_NickServ, source, ns_servhelper_help);
	if (is_oper(source))
	        notice_list(s_NickServ, source, ns_servircop_help);
	if (is_services_oper(source))
	        notice_list(s_NickServ, source, ns_servop_help);
	if (is_services_admin(source))
	        notice_list(s_NickServ, source, ns_servadmin_help);
	if (is_services_root(source))
	        notice_list(s_NickServ, source, ns_servroot_help);
    }

    notice(s_NickServ, source,
       "\2END OF OHELP\2");

}

static void do_help(const char *source)
{
    char *cmd = strtok(NULL, " ");
//    char buf[BUFSIZE];

    if (!cmd) {
       notice(s_NickServ, source, "\2NickServ\2 allows you to \"register\" a nickname and");
       notice(s_NickServ, source, "prevent others from using them.  If the nick is not used");
       notice(s_NickServ, source, "for %d days, the registration will expire.  The following", NICK_EXPIRE);
       notice(s_NickServ, source, "commands allow for registration and maintenance of");
       notice(s_NickServ, source, "nicknames; to use them, type \2/msg NickServ <command>\2.");
       notice(s_NickServ, source, "For more information on a specific command, type");
       notice(s_NickServ, source, "\2/msg NickServ help <command>\2.");
       notice_list(s_NickServ, source, ns_help_index);
       if (is_services_helpop(source))
       notice(s_NickServ, source, "\2IRCop\2: please use \2OHELP\2 for more commands.");
    } else if (stricmp(cmd, "REGISTER") == 0) {
       notice_list(s_NickServ, source, ns_help_register);
    } else if (stricmp(cmd, "IDENTIFY") == 0) {
       notice_list(s_NickServ, source, ns_help_identify);
    } else if (stricmp(cmd, "ACCESS") == 0) {
       notice_list(s_NickServ, source, ns_help_access);
    } else if (stricmp(cmd, "SET") == 0) {
	    char *cmd1 = strtok(NULL, " ");
	    if (!cmd1) {
	       notice_list(s_NickServ, source, ns_help_set);
	    } else if (stricmp(cmd1, "KILL") == 0) {
	       notice_list(s_NickServ, source, ns_help_set_kill);
	    } else if (stricmp(cmd1, "SECURE") == 0) {
	       notice_list(s_NickServ, source, ns_help_set_secure);
	    } else if (stricmp(cmd1, "PRIVATE") == 0) {
	       notice_list(s_NickServ, source, ns_help_set_private);
	    } else if (stricmp(cmd1, "PASSWD") == 0) {
	       notice_list(s_NickServ, source, ns_help_set_password);
	    } else if (stricmp(cmd1, "PASSWORD") == 0) {
	       notice_list(s_NickServ, source, ns_help_set_password);
	    } else if (stricmp(cmd1, "EMAIL") == 0) {
	       notice_list(s_NickServ, source, ns_help_set_email);
	    } else if (stricmp(cmd1, "HIDEMAIL") == 0) {
	       notice_list(s_NickServ, source, ns_help_set_hidemail);
	    } else if (stricmp(cmd1, "NOMEMO") == 0) {
	       notice_list(s_NickServ, source, ns_help_set_nomemo);
	    } else if (stricmp(cmd1, "NEVEROP") == 0) {
	       notice_list(s_NickServ, source, ns_help_set_neverop);
	    } else if (stricmp(cmd1, "EMAILMEMOS") == 0) {
	       notice_list(s_NickServ, source, ns_help_set_emailmemos);
	    } else if (stricmp(cmd1, "PRIVMSG") == 0) {
	       notice_list(s_NickServ, source, ns_help_set_privmsg);
	    } else if (stricmp(cmd1, "AGE") == 0) {
	       notice_list(s_NickServ, source, ns_help_set_age);
	    } else if (stricmp(cmd1, "SEX") == 0) {
	       notice_list(s_NickServ, source, ns_help_set_sex);
	    } else if (stricmp(cmd1, "UIN") == 0) {
	       notice_list(s_NickServ, source, ns_help_set_uin);
	    } else if (stricmp(cmd1, "URL") == 0) {
	       notice_list(s_NickServ, source, ns_help_set_url);
	    } else if (stricmp(cmd1, "INFO") == 0) {
	       notice_list(s_NickServ, source, ns_help_set_info);
	    } else {
	       notice_list(s_NickServ, source, ns_help_nohelp); 
	    }
    } else if (stricmp(cmd, "DROP") == 0) {
       notice_list(s_NickServ, source, ns_help_drop);
    } else if (stricmp(cmd, "RECOVER") == 0) {
       notice_list(s_NickServ, source, ns_help_recover);
    } else if (stricmp(cmd, "RELEASE") == 0) {
       notice_list(s_NickServ, source, ns_help_release);
    } else if (stricmp(cmd, "ACC") == 0) {
       notice_list(s_NickServ, source, ns_help_acc);
    } else if (stricmp(cmd, "LINK") == 0) {
       notice_list(s_NickServ, source, ns_help_link);
    } else if (stricmp(cmd, "GHOST") == 0) {
       notice_list(s_NickServ, source, ns_help_ghost);
    } else if (stricmp(cmd, "INFO") == 0) {
       notice_list(s_NickServ, source, ns_help_info);
    } else if (stricmp(cmd, "LISTCHANS") == 0) {
       notice_list(s_NickServ, source, ns_help_listchans);
    } else if (stricmp(cmd, "SELFDEL") == 0) {
       notice_list(s_NickServ, source, ns_help_selfdel);
    } else if (stricmp(cmd, "LINKS") == 0) {
       notice_list(s_NickServ, source, ns_help_links);
    } else if (stricmp(cmd, "UNLINK") == 0) {
       notice_list(s_NickServ, source, ns_help_unlink);
    } else if (stricmp(cmd, "WHOIS") == 0) {
       notice_list(s_NickServ, source, ns_help_whois);
    } else if (stricmp(cmd, "VERIFY") == 0) {
       notice_list(s_NickServ, source, ns_help_verify);
    } else if (stricmp(cmd, "NICK") == 0) {
       notice_list(s_NickServ, source, ns_help_nick);
    } else if (stricmp(cmd, "WHOIAM") == 0) {
       notice_list(s_NickServ, source, ns_help_whoiam);
    } else {
       notice_list(s_NickServ, source, ns_help_nohelp); 
    }

/*
    snprintf(buf, sizeof(buf), "%s%s", s_NickServ, cmd ? " " : "");
    strscpy(buf+strlen(buf), cmd ? cmd : "", sizeof(buf)-strlen(buf));
    helpserv(s_NickServ, source, buf);
*/

}

/*************************************************************************/

/* Register a nick. */

static void do_register(const char *source)
{
    NickInfo *ni;
    User *u;
    char *gender;
    char *pass = strtok(NULL, " ");
    char *email = strtok(NULL, " ");
    int i, j;
    long int guestbuf = 0;
    double x = RAND_MAX + 1.0;
    int y;
    FILE *fp;
    char cmdbuf[500];
    char cmdbuf2[500];
    char newcode[64];
    y = 1 + rand() * (99999 / x);
    guestbuf = y;
    snprintf (newcode, sizeof (newcode), "auspice%d", guestbuf);
    if (readonly) {
	notice(s_NickServ, source,
		"Sorry, nickname registration is temporarily disabled.");
	return;
    }

    if (!pass || !email) {

        notice(s_NickServ, source, "Syntax: \2REGISTER\2 <password> <email>"); 
	notice(s_NickServ, source,
			"\2/msg %s HELP REGISTER\2 for more information.",
			s_NickServ);

    } else if (strlen(pass) >= PASSMAX) {

        notice(s_NickServ, source, "Password too long."); 

    } else if (!(u = finduser(source))) {

	log("%s: Can't register nick %s: nick not online", s_NickServ, source);
	notice(s_NickServ, source, NS_REGFAIL);

    } else if (time(NULL) < u->lastnickreg + NS_REG_DELAY) {

	notice(s_NickServ, source,
	    "Please wait %d seconds before using the REGISTER command again.",
	    NS_REG_DELAY);

    } else if (ni = findnick(source)) {

	if (ni->flags & NI_VERBOTEN) {
	    log("%s: %s@%s tried to register FORBIDden nick %s", s_NickServ,
			u->username, u->host, source);
	    notice(s_NickServ, source,
			"Nickname \2%s\2 may not be registered.", source);
	} else {
	    notice(s_NickServ, source,
			"Nickname \2%s\2 is already registered!", source);
	}

    } else if (stricmp(source, pass) == 0 || strlen(pass) < 5) {

	notice(s_NickServ, source,
		"Please try again with a more obscure password.  Passwords "
		"should be at least five characters long, should not be "
		"something easily guessed (e.g. your real name or your "
		"nick), and cannot contain the space or tab characters.  "
		"\2/msg %s HELP REGISTER\2 for more information.", s_NickServ);

    } else if (match_wild_nocase("guest*", source)) {
         notice(s_NickServ, source,
                "You may not register nicks starting with \"Guest\". "
                "\2/msg %s HELP REGISTER\2 for more information.", s_NickServ);

    } else if (strlen(pass) > PASSMAX) {
         notice(s_NickServ, source,
             "Sorry, nick passwords may only be up to %d characters long",
                 PASSMAX);

     } else if (!strchr(email, '@')) {
            notice(s_NickServ, source,
               "ERROR: E-Mail Address must contain an '@'");
     } else if (!strchr(email, '.')) {
            notice(s_NickServ, source,
               "ERROR: Look like you give invalid email.");
     } else if (strchr(email, '\\')) {
           notice(s_NickServ, source,
               "ERROR: invalid email characters.");
    } else {
    if(!is_netadmin(source))
    for (j=1; j <= nscount; j++) {
	if (stricmp(bannick[j], source) == 0) {
	        notice(s_NickServ, source, NS_CANNOTREG); 
		return;
	}
    }
	if (ni = makenick(source)) {
	    strscpy(ni->pass, pass, PASSMAX);
	    if (VALIDEEMAIL == 0) {
		    ni->flags = NI_IDENTIFIED | NI_RECOGNIZED;
	            if (!is_on_id_list(source, source)) {
        	        for (i=0; u->id_nicks[i] && i < MAX_IDS; i++)
                	    ;
	                if (i < MAX_IDS)
        	             u->id_nicks[i] = sstrdup(source);
	            }
            }

	    ni->flags |= NI_MEMO_SIGNON | NI_MEMO_RECEIVE;
	    ni->last_usermask = smalloc(strlen(u->username)+strlen(u->host)+2);
/*
            if (AUTO_NOHOST > 0)
                ni->flags |= NI_NOHOST; 
*/
	    if (is_oper(source))
	   	ni->flags |= NI_HOP;

	    if (NS_PRIVATE > 0)
	   	ni->flags |= NI_PRIVATE;
	    if (NS_KILL > 0)
		ni->flags |= NI_KILLPROTECT;
	    if (NS_SECURE > 0)
		ni->flags |= NI_SECURE;

	    sprintf(ni->last_usermask, "%s@%s", u->username, u->host);
	    ni->last_realname = sstrdup(u->realname);
	    ni->time_registered = ni->last_seen = time(NULL);
	    ni->email = sstrdup (email);
	    ni->accesscount = 1;
            ni->news = 0;
            ni->uin = NULL;
            ni->age = NULL;
            ni->info = NULL;
            ni->sex = NULL;
            ni->mlock = NULL;
//	    ni->auth = sstrdup(newcode);
	    ni->access = smalloc(sizeof(char *));
	    ni->access[0] = create_mask(u);
            ni->memomax = DEF_MAX_MEMOS;
	    clear_memos(ni->nick);
	    log("%s: `%s' registered by %s@%s", s_NickServ,
		    source, u->username, u->host);
            slog("NS R %s (%s@%s)", source, u->username, u->host);
            do_break_log("NS_R", "NS R %s (%s@%s)",
                 source, u->username, u->host);

	    if (VALIDEEMAIL == 0) {
		    time_t now = time(NULL);
        	    notice(s_NickServ, source, NS_REGISTER, source, ni->access[0]);
	            notice(s_NickServ, source, NS_REGISTER2, pass);
        	    notice(s_NickServ, source, NS_REGMAIL, email);
	 	    notice(s_NickServ, source, NS_KSECRET);
		    notice(s_NickServ, source, NS_HIDEMAILC, s_NickServ);
	            notice(s_NickServ, source, MISC_OFHELP, AOCHANS);
		    notice(s_NickServ, source, NS_INFOOWNNICK);
		    notice(s_NickServ, source, "http://%s", nsurl);
		    if (is_oper(source) && !(ni->flags & NI_HOP))
		        notice(s_NickServ, source, NS_TURNIRCOP);
	    } else {
		notice(s_NickServ, source, "Your nick (%s) isn't activate yet.",source);
		notice(s_NickServ, source, "Type /msg NickServ AUTH %s <code>",source);
		notice(s_NickServ, source, "The Authorize code have been sent to your e-mail after you have registered.");
		ni->flags |= NI_WAUTH;
	    }

	    if (REGNOTICE == 1)
	    wallops(s_NickServ, "Nick \2%s\2 is registered to %s",u->nick, ni->access[0]);
/*	    
	    if (VALIDEEMAIL == 1) {
	                fp = fopen("memo.txt", "a");
        	        fprintf (fp, "From: %s <%s>\n", s_NickServ, PASS_EMAIL);
	                fprintf (fp, "To: %s <%s>\n", ni->nick, ni->email);
        	        fprintf (fp, "Subject: Authorize code!!\n");
	                fprintf (fp, "You received this email because you have registered a nickname\n");
	                fprintf (fp, "in %s IRC Network.", NETWORK_NAME);
	                fprintf (fp, "Your nickname (%s) authorize code is %s\n\n",ni->nick, ni->auth);
	                fprintf (fp, "Type /msg NickServ Auth %s %s\n\n",ni->nick, ni->auth);
			fprintf (fp, "If you leave this nick Unauthorize for one day this nick will expire.\n");
		        fclose (fp);

			sprintf (cmdbuf, "%s -t %s < %s", sendmail_path, ni->email, "memo.txt");
			sprintf (cmdbuf2, "rm -f %s", "memo.txt");

	              system(cmdbuf);
	              system(cmdbuf2);
		}

*/
#if REG_SAVES > 0
            if (regcnt++ >= REG_SAVES) {
                wallops(SERVER_NAME,
                     "Automatic DataBase save on %d new registrations",
                          regcnt);
                save_data = 1;
            }
#endif

            show_next_db(source, s_NickServ);
	    u->lastnickreg = time(NULL);

	    if (VALIDEEMAIL == 0)
	            send_cmd(s_NickServ, "SVSMODE %s +r 0", source);

	    if (FIRSTUSED == 0)
		    do_sendnreg(source, create_mask(u), pass, email);
	} else {
	    log("%s: makenick(%s) failed", s_NickServ, source);
	    notice(s_NickServ, source, "Sorry, couldn't register your nickname.");

	}
	}	
}

/*************************************************************************/
static void do_identify(const char *source)
{
    char *nick = strtok(NULL, " ");
    char *pass = strtok(NULL, " ");
    char *tmp;
    NickInfo *ni, *hni;
    User *u;
    int res;
    char access[128];
    int i, x=0;
    int j;
    int is_force;
    
    *access = 0;

    if (!nick) {
        notice(s_NickServ, source, "Syntax: \2IDENTIFY\2 [nick] <password>");
	notice(s_NickServ, source,
		"\2/msg %s HELP IDENTIFY\2 for more information.",
		s_NickServ);
	return;
    }

    if (!pass) {
	pass = sstrdup(nick);
	nick = sstrdup(source);
	x = 1;
    }

    ni = findnick(nick);
    hni = host(ni);

    is_force = (is_services_root(source) && !stricmp(pass, "force"));

    if (!ni || !hni) {
	notice(s_NickServ, source, "Your nick (%s) isn't registered.",nick);

    } else if (!(u = finduser(source))) {

	log("%s: IDENTIFY from nonexistent nick %s", s_NickServ, source);
	notice(s_NickServ, source, NS_IDFAIL);

    } else if (hni->flags & NI_WAUTH) {
	notice(s_NickServ, source, "Your nick (%s) isn't activate yet.",nick);
	notice(s_NickServ, source, "Type /msg NickServ AUTH %s <code>",nick);
	notice(s_NickServ, source, "The Authore code have been sent to your e-mail after you have registered.");
    } else if (u->flags & U_ABUSIVE) {

	log("%s: Failed identify from abusive user %s", s_NickServ, source);
	notice(s_NickServ, source, "Sorry, You have the abusive flag set");

      if (EXTRA_SNOOP > 0) {
          slog("NS *I %s (%s!%s@%s) [Abusive]",
		nick, source, u->username, u->host);
          do_break_log("NS_I", "NS *I %s (%s!%s@%s)",
              nick, source, u->username, u->host);
      }

    } else if (!is_force && !(res = check_password(pass, hni->pass))) {

	log("%s: Failed IDENTIFY for %s (%s!%s@%s)",
		s_NickServ, nick, source, u->username, u->host);
        if (EXTRA_SNOOP) {
            slog("NS *I %s (%s!%s@%s)",
	         nick, source, u->username, u->host);
            do_break_log("NS_I", "NS *I %s (%s!%s@%s)",
                 nick, source, u->username, u->host);
        }
        notice(s_NickServ, source, NS_BADPASS);
	bad_password(u, nick);

    } else if (res == -1) {

	notice(s_NickServ, source, NS_IDFAIL);

    } else {

        if (is_autokilled(source)) {
	  wallops(s_NickServ, "\2Warning\2 - Found identified nick \2%s\2 on AUTO KILL list", source);
          slog("NS A*I %s (%s!%s@%s)",
                source, source, u->username, u->host);
          do_break_log("NS_I", "NS A*I %s (%s!%s@%s)",
               source, source, u->username, u->host);
          kill_user(SERVER_NAME, source, "You are Auto-KILLED from this  network, please stay away!");
          return;
       }

     if (ni->flags & NI_FREEZE) {
        notice(s_NickServ, source, "This nick name can not Identify as it frozed.");
	return;	
     }

      if ((!(ni->flags & NI_IDENTIFIED)) && (EXTRA_SNOOP > 0)) {
          slog("NS I %s (%s!%s@%s) %s",
		nick, source, u->username, u->host,
                is_force ? "[Forced]" : "");
          do_break_log("NS_I", "NS I %s (%s!%s@%s) %s",
		nick, source, u->username, u->host,
                is_force ? "[Forced]" : "");
       }

     if (is_hostline(source))
	do_hostline(source);

     if (ni->flags & NI_IDENTIFIED) {
        notice(s_NickServ, source, NS_ALREADYID);
	if (!(ni->flags & NI_NEVEROP))
		do_doop(source);
	return;	
     }

	ni->flags |= NI_IDENTIFIED;
	ni->id_timestamp = u->signon;

	if (!is_on_id_list(u->nick, hni->nick)) {
	    for (i=0; u->id_nicks[i] && i < MAX_IDS; i++)
		;
	    if (i < MAX_IDS)
	         u->id_nicks[i] = sstrdup(hni->nick);
	}

	if (!(ni->flags & NI_RECOGNIZED)) {
	    ni->last_seen = time(NULL);

	    if (hni->last_usermask)
		free(hni->last_usermask);
	    hni->last_usermask = smalloc(strlen(u->username)+strlen(u->host)+2);
	    sprintf(hni->last_usermask, "%s@%s", u->username, u->host);
	    if (ni->last_realname)
		free(ni->last_realname);
	    ni->last_realname = sstrdup(u->realname);
            if (!ni->news)   // If nick registered after news implement
                ni->news = 0;
            if (!ni->uin)
                ni->uin = NULL;
            if (!ni->age)
                ni->age = NULL;
            if (!ni->info)
                ni->info = NULL;
            if (!ni->sex)
                ni->sex = NULL; 
            if (!ni->mlock)
                ni->mlock = NULL; 
/*
            if (!ni->auth)
                ni->auth = NULL; 
*/
	    log("%s: %s!%s@%s identified for nick %s %s", s_NickServ,
			source, u->username, u->host, source,
                        is_force ? "[Forced]" : "");
	}

	if (is_services_coder(source)) {
                if (ENABLE_UMODE_a)
                    send_cmd(s_NickServ,"SVSMODE %s +aqS 0",source);
		    send_cmd(s_NickServ,"SWHOIS %s :is a Services Master",u->nick);
	            strcpy(access, "as a Services Master.");
		    m_smartinfo(source);
	} else if (is_services_root(source)) {
                if (ENABLE_UMODE_a)
                    send_cmd(s_NickServ,"SVSMODE %s +aqS 0",source);
		    send_cmd(s_NickServ,"SWHOIS %s :is a Services Root",u->nick);
                strcpy(access, "as a Services Root.");
		    m_smartinfo(source);
	} else if (is_services_admin(source)) {
                if (ENABLE_UMODE_a)
                    send_cmd(s_NickServ,"SVSMODE %s +aqS 0",source);
		    send_cmd(s_NickServ,"SWHOIS %s :is a Services Admin",u->nick);
	            strcpy(access, "as a Services Admin.");
		    m_smartinfo(source);
	} else if (is_services_oper(source) && !is_servadmin(source)) {
		wallops(NULL,"%s now is a Services Operator", u->nick);
                send_cmd(s_NickServ,"SVSMODE %s +hwefbcgswW 0",source);
                strcpy(access, "as a Services Operator.");
	} else if (is_services_helpop(source) && !is_oper(source)) {
//		wallops(NULL,"%s now is a Services Helper", u->nick);
                send_cmd(s_NickServ,"SVSMODE %s +hwefbcgswW 0",source);
                strcpy(access, "as a Services HelpOp.");
	} else {
                strcpy(access, "");
	}

	if (!(ni->flags & NI_NEVEROP))
		do_doop(source);

	if (ni->mlock) {
	        send_cmd(s_NickServ,"SVSMODE %s %s 0",source, ni->mlock);
	        notice(s_NickServ, source, "Your User mode auto set to \2%s\2", ni->mlock);
	}

        notice(s_NickServ, source, NS_IDENTIFY,ni->nick,access);
            send_cmd(s_NickServ, "SVSMODE %s +r 0", source);

	if (is_oper(source) && !(ni->flags & NI_HOP))
	        notice(s_NickServ, source, NS_TURNIRCOP);

	if (hni->flags & NI_EXMAIL) {
	       hni->flags &= ~NI_EXMAIL;
	       notice(s_NickServ, source, "Welcome Back!! Long time no see, how are you?.");
	}

	if (!(ni->flags & NI_RECOGNIZED))
	    check_memos(source);
    }

    if (x) {
      free(nick);
      free(pass);
    }

}

/*************************************************************************/

static void do_delete(const char *source)
{
    char *nick = strtok(NULL, " ");
    NickInfo *ni;
    User *u = finduser(source);
    int is_root = is_services_root(source);

    if (readonly) {
	notice(s_NickServ, source,
                "Sorry, nickname deletion is temporarily disabled. (Read Only)");
	return;
    }
    if (!nick) {
       notice(s_NickServ, source, "\2Delete\2: You MUST specify a nick");
       return;
    }

    if (!(ni = findnick(nick))) {
        if (nick)
           notice(s_NickServ, source, NS_NOTREG, nick);
        return;
    }
    if (stricmp(nick, SERVICES_MASTER) == 0 && !is_services_coder(source)) {
        notice(s_NickServ, source,
            "You are not permitted to delete a Master's account");
        return;
    }
    del_hostline(ni->nick);
    do_expireanick(ni->nick);
    delnick(ni);
    slog("NS +D %s (%s!%s@%s)", nick, source, u->username, u->host);
    do_break_log("NS_G", "NS +D %s (%s!%s@%s)",
       nick, source, u->username, u->host);
    log("%s: %s!%s@%s deleted nickname %s", s_NickServ,
            source, u->username, u->host, nick);
        notice(s_NickServ, source, NS_DELETE, nick);
        wallops(s_NickServ, "%s deleted nickname \2%s\2", source, nick);
}



/*************************************************************************/
NickInfo * host (NickInfo * ni)
{
    if (!ni)
        return NULL;
    if (ni->flags & NI_SLAVE)
        return findnick (ni->last_usermask);
    else
        return ni;
}
/*************************************************************************/
int
userisnick (const char *nick)
{
    NickInfo *ni = findnick (nick);
    if (!finduser (nick) || !ni)
        return 0;
    if ((getflags (ni) & NI_SECURE) ? (ni->flags & NI_IDENTIFIED)
        : (ni->flags & (NI_RECOGNIZED | NI_IDENTIFIED)))
        return 1;
    return 0;
}
/*************************************************************************/
/* Find flags (assume host's flags if there is one) */

long
getflags (NickInfo * ni)
{
    NickInfo *hni = host (ni);
    long flags;

    if (hni)
    {
        flags = hni->flags;
        flags &= ~NI_RECOGNIZED;
        flags &= ~NI_IDENTIFIED;
        flags &= ~NI_KILL_HELD;
        if (ni->flags & NI_RECOGNIZED)
            flags |= NI_RECOGNIZED;
        if (ni->flags & NI_IDENTIFIED)
            flags |= NI_IDENTIFIED;
        if (ni->flags & NI_KILL_HELD)
            flags |= NI_KILL_HELD;
        if (ni->flags & NI_SLAVE)
            flags |= NI_SLAVE;
        return flags;
    }
    return 0;
}

/*************************************************************************/
NickInfo * slave(const char *nick, int num)
{
    NickInfo *ni;
    int i, cnt = 0;

    for (i = 33; i < 256; ++i)
        for (ni = nicklists[i]; ni; ni = ni->next)
        {
            if (stricmp (ni->last_usermask, nick) == 0)
                cnt++;
            if (cnt == num)
                return ni;
        }
    return NULL;
}
/*************************************************************************/
int
issibling (NickInfo * ni, const char *target)
{
    if (ni)
    {
        NickInfo *hni = host (ni);
        int i;
        if (stricmp (ni->nick, target) == 0)
            return 1;
        if (stricmp (hni->nick, target) == 0)
            return 1;
        for (i = slavecount (hni->nick); i; i--)
            if (stricmp (target, slave (hni->nick, i)->nick) == 0)
                return 1;
    }
    return 0;
}
/*************************************************************************/
int slavecount (const char *nick)
{
    NickInfo *ni;
    int i, cnt = 0;

    for (i = 33; i < 256; ++i)
        for (ni = nicklists[i]; ni; ni = ni->next)
            if (stricmp (ni->last_usermask, nick) == 0)
                cnt++;
    return cnt;
}
/*************************************************************************/
static void do_slaves (const char *source)
{
    char *nick = strtok (NULL, " ");
    NickInfo *ni;
    User *u = finduser (source);

    if (!is_oper (source) && nick)
    {
        notice (s_NickServ, source, "Syntax: \2LINKS\2%s", is_oper(source) ? " [Nick]" : "");
        notice (s_NickServ, source, ERR_MORE_INFO, s_NickServ, "LINKS");
    }
    else if (!(ni = findnick (nick ? nick : source)))
        if (nick)
            notice (s_NickServ, source, NS_NOT_REGISTERED, nick);
        else
            notice (s_NickServ, source, NS_YOU_NOT_REGISTERED, source);

    else if (!u)
        notice (s_NickServ, source, ERR_YOU_DONT_EXIST);

    else if (!nick && !userisnick(source))
    {
        notice (s_NickServ, source, NS_NOT_YOURS, source);
        notice (s_NickServ, source, ERR_IDENT_NICK_FIRST, s_NickServ);
        if (!u)
            log ("[ERR] %s: LINKS from nonexistent user %s", s_NickServ, source);
    }
    else
    {
        NickInfo *hni = host(ni);
        int i = slavecount (hni->nick);
        char buf[NICKMAX * i];
        char buf2[256];

        if (i < 1)
            if (nick)
                notice (s_NickServ, source, NS_NO_SLAVES, nick);
            else
                notice (s_NickServ, source, NS_YOU_NO_SLAVES);
        else
        {
            *buf = 0;
            sprintf (buf, "\2%s\2: ", hni->nick);
            for (; i > 0; i--) {
                sprintf (buf2, "%s ", slave (hni->nick, i)->nick);
                strcat (buf, buf2);
                if (i==1)
                    notice (s_NickServ, source, "%s", buf);
            }
        }
    }
}

/*************************************************************************/
static void do_drop(const char *source)
{
    char *nick = strtok(NULL, " ");
    NickInfo *ni;
    User *u = finduser(source);
    int i;

    if (readonly) {
	notice(s_NickServ, source,
		"Sorry, nickname de-registration is temporarily disabled.");
	return;
    }

    if (!nick) {
       notice(s_NickServ, source, "Syntax: \2DROP\2 <nick>");
       return;
    }

    if (!(ni = findnick(nick))) {
	    notice(s_NickServ, source, "Your nick isn't registered.");

    } else if (!is_on_id_list(source, nick)) {

        slog("NS *D %s (%s!%s@%s) [User Hasn't Identified]", source, source, u->username, u->host);
        do_break_log("NS_O", "NS *D %s (%s!%s@%s) [User Hasn't Identified]", source, source, u->username, u->host);
	notice(s_NickServ, source,
		"Password authentication required for that command.");
	notice(s_NickServ, source,
		"Retry after typing \2/msg %s IDENTIFY [\37nick\37] \37password\37.",
		s_NickServ);
	if (!u)
	    log("%s: DROP from nonexistent user %s", s_NickServ, source);

    } else if (!u) {

	log("%s: DROP %s from nonexistent oper %s", s_NickServ, nick, source);
	notice(s_NickServ, source, "Can't find your user record!");

    } else {

	if (readonly) {
	    notice(s_NickServ, source,
		"Warning: Services is in read-only mode.  Changes will not be saved.");
	}
        if (stricmp(nick, SERVICES_MASTER) == 0 && !is_services_coder(source)) {
            notice(s_NickServ, source,
               "You are not permitted to delete a Master's account");
            return;
        }

        do_expireanick(ni->nick);
	del_hostline(ni->nick);
	delnick(ni);

        slog("NS D %s (%s!%s@%s)", nick, source, u->username, u->host);
        do_break_log("NS_O", "NS D %s (%s!%s@%s)",
             nick, source, u->username, u->host);
	log("%s: %s!%s@%s dropped nickname %s", s_NickServ,
		source, u->username, u->host, nick ? nick : source);
        notice(s_NickServ, source, NS_DROP);
    }
}

/*************************************************************************/

static void do_set(const char *source)
{
    char *cmd    = strtok(NULL, " ");
    char *param  = strtok(NULL, "");
    NickInfo *ni = findnick(source);
    User *u;

    if (readonly) {
	notice(s_NickServ, source,
		"Sorry, nickname option setting is temporarily disabled.");
	return;
    }

    if (!param) {

	notice(s_NickServ, source,
                "Syntax: \2SET\2 <command> [<parameters>]");
	notice(s_NickServ, source,
		"\2/msg %s HELP SET\2 for more information.", s_NickServ);

    } else if (!(ni = findnick(source))) {

	notice(s_NickServ, source, NS_NOTREGED);

    } else if (!(u = finduser(source)) || !(ni->flags & NI_IDENTIFIED)) {

	notice(s_NickServ, source,
		"Password authentication required for that command.");
	notice(s_NickServ, source,
		"Retry after typing \2/msg %s IDENTIFY \37password\37.",
		s_NickServ);
	if (!u)
	    log("%s: SET from nonexistent user %s", s_NickServ, source);

    } else if (stricmp(cmd, "PASSWORD") == 0) {

	do_set_password(ni, param);

    } else if (stricmp(cmd, "PASSWD") == 0) {

	do_set_password(ni, param);

    } else if (stricmp(cmd, "URL") == 0) {

	do_set_url(ni, param);

    } else if (stricmp(cmd, "EMAIL") == 0) {

	do_set_email(ni, param);

    } else if (stricmp(cmd, "UIN") == 0) {

	do_set_uin (ni, param);

//    } else if (stricmp(cmd, "MLOCK") == 0) {
	
//	do_set_mlock (ni, param);

    } else if (stricmp(cmd, "AGE") == 0) {

	do_set_age (ni, param);

    } else if (stricmp(cmd, "INFO") == 0) {

	do_set_info (ni, param);

    } else if (stricmp(cmd, "SEX") == 0) {

	do_set_sex (ni, param);

    } else if (stricmp(cmd, "KILL") == 0) {

	do_set_kill(ni, param);

    } else if (stricmp(cmd, "SECURE") == 0) {

	do_set_secure(ni, param);

    } else if (stricmp(cmd, "PRIVATE") == 0) {

	do_set_private(ni, param);

    } else if (stricmp(cmd, "EMAILMEMOS") == 0) {
	do_set_emailmemos(ni, param);

    } else if (stricmp(cmd, "NEVEROP") == 0) {
        do_set_neverop(ni, param);

    } else if (stricmp(cmd, "NOOP") == 0) {
        do_set_noop(ni, param);

    } else if (stricmp(cmd, "AUTH") == 0) {
        do_set_auth(ni, param);

    } else if (stricmp(cmd, "MALAY") == 0) {
        do_set_malay(ni, param);

    } else if (stricmp(cmd, "PRIVMSG") == 0) {
        do_set_privmsg(ni, param);

    } else if (stricmp(cmd, "NOMEMO") == 0) {
        do_set_nomemo(ni, param);

    } else if (stricmp(cmd, "NOHOST") == 0) {
        do_set_nohost(ni, param);

    } else if (stricmp(cmd, "HIDEMAIL") == 0) {
	do_set_hidemail(ni, param);
    } else if (stricmp(cmd, "BOTMSG") == 0) {
	do_set_botmsg(ni, param);
    } else if (stricmp(cmd, "IRCOP") == 0) {
	if (is_oper(source))
		do_set_ircop(ni, param);

    } else {
	notice(s_NickServ, source,
			"Unknown SET option \2%s\2.", strupper(cmd));
    }
}

/*************************************************************************/

static void do_set_password(NickInfo *ni, char *param)
{
    NickInfo *hni = host(ni);
    char *source = ni->nick;
    if (strlen(param) >= PASSMAX) {
	    notice(s_NickServ, ni->nick, "Password too long.");
    } else {
	    strscpy(hni->pass, param, PASSMAX);
	    notice(s_NickServ, ni->nick, "Password changed to \2%s\2.", param);
	    show_next_db(source, s_NickServ);
    }
}

/*************************************************************************/

static void do_set_url(NickInfo *ni, char *param)
{
    NickInfo *hni = host(ni);
    char *source = ni->nick;
    if (hni->url)
	free(hni->url);

    if (stricmp(param,"none")) {
        hni->url = sstrdup(param);
        notice(s_NickServ, ni->nick,
		"URL for %s changed to \2%s\2.", ni->nick, param);
    } else {
       if (!hni->url) {
            notice(s_NickServ, ni->nick,
                "You have no url set to disable!");
            return;
       }

       ni->url = NULL;
       notice(s_NickServ, source, "URL disabled for nick \2%s\2", ni->nick);
    }
    show_next_db(source, s_NickServ);
}

/*************************************************************************/

static void do_set_email(NickInfo *ni, char *param)
{
    NickInfo *hni = host(ni);
    char *source = ni->nick;

    if ((stricmp(param,"none")) && (stricmp(param,"hide"))) {

       if (!strchr(param, '@')) {
            notice(s_NickServ, source,
               "ERROR: E-Mail Address must contain an '@'");
            return;
       }
       if (strchr(param, '\\')) {
           notice(s_NickServ, source,
               "ERROR: invalid email characters.");
           return;
       }

       if (hni->email)
           hni->email = NULL;

       hni->email = sstrdup(param);
       notice(s_NickServ, ni->nick,
		"E-mail address for %s changed to \2%s\2.", ni->nick, param);
       slog("NS E %s [%s]", ni->nick, param);

    } else if (!stricmp(param,"hide")) {
         if (!hni->email) {
            notice(s_NickServ, ni->nick,
                "You have no email address set to disable!");
            return;
         }
         hni->flags |= NI_HIDE_EMAIL;
         notice(s_NickServ, ni->nick,
                  "Your email address will no longer be shown");
         show_next_db(source, s_NickServ);
         return;

    } else {
       if (!hni->email) {
           notice(s_NickServ, ni->nick,
               "You have no email address set to disable!");
           return;
       }

       hni->flags &= ~NI_HIDE_EMAIL;
       hni->email = NULL;
       notice(s_NickServ, source, "EMAIL disabled for nick \2%s\2", ni->nick);
    }
    show_next_db(source, s_NickServ);
}

/*************************************************************************/

static void do_set_kill(NickInfo *ni, char *param)
{
    NickInfo *hni = host(ni);
    char *source = ni->nick;

    if (stricmp(param, "ON") == 0) {

	hni->flags |= NI_KILLPROTECT;
	notice(s_NickServ, source, "Kill protection is now \2ON\2.");
        show_next_db(source, s_NickServ);

    } else if (stricmp(param, "OFF") == 0) {

	hni->flags &= ~NI_KILLPROTECT;
	notice(s_NickServ, source, "Kill protection is now \2OFF\2.");
        show_next_db(source, s_NickServ);

    } else {

        notice(s_NickServ, source, "Syntax: \2SET KILL\2 [\37ON\37|\37OFF\37]");
	notice(s_NickServ, source,
		"\2/msg %s HELP SET KILL\2 for more information.", s_NickServ);
    }
}

/*************************************************************************/

static void do_set_secure(NickInfo *ni, char *param)
{
    NickInfo *hni = host(ni);
    char *source = ni->nick;

    if (stricmp(param, "ON") == 0) {

	hni->flags |= NI_SECURE;
	notice(s_NickServ, source, "Secure option is now \2ON\2.");
        show_next_db(source, s_NickServ);

    } else if (stricmp(param, "OFF") == 0) {

	hni->flags &= ~NI_SECURE;
	notice(s_NickServ, source, "Secure option is now \2OFF\2.");
        show_next_db(source, s_NickServ);

    } else {

        notice(s_NickServ, source, "Syntax: \2SET SECURE\2 [\37ON\37|\37OFF\37]");
	notice(s_NickServ, source,
		"\2/msg %s HELP SET SECURE\2 for more information.",
		s_NickServ);
    }
}

/*************************************************************************/

static void do_set_private(NickInfo *ni, char *param)
{
    NickInfo *hni = host(ni);
    char *source = ni->nick;

    if (stricmp(param, "ON") == 0) {

	hni->flags |= NI_PRIVATE;
	notice(s_NickServ, source, "Private option is now \2ON\2.");
        show_next_db(source, s_NickServ);

    } else if (stricmp(param, "OFF") == 0) {

	hni->flags &= ~NI_PRIVATE;
	notice(s_NickServ, source, "Private option is now \2OFF\2.");
        show_next_db(source, s_NickServ);

    } else {

        notice(s_NickServ, source, "Syntax: \2SET PRIVATE\2 [\37ON\37|\37OFF\37]");
	notice(s_NickServ, source,
		"\2/msg %s HELP SET PRIVATE\2 for more information.",
		s_NickServ);
    }
}

static void do_set_emailmemos(NickInfo *ni, char *param)
{
    NickInfo *hni = host(ni);
    char *source = ni->nick;


    if (stricmp(param, "ON") == 0) {
	if (!hni->email) {
	    notice(s_NickServ,source,"You must set your e-mail address.");
	    notice(s_NickServ,source,"Do do this type: \2/msg %s set email \37email\37\2",s_NickServ);
	    return;
        } else if ((ALLOW_EMAIL == 0) && !is_services_root(source)) {
              notice(s_NickServ, source,
                    "Sorry, emailing memos have been disabled");
              return;
        }
        hni->flags |= NI_EMAILMEMOS;
        notice(s_NickServ, source, "Your memos will now be emailed to %s.", ni->email);
        show_next_db(source, s_NickServ);

    } else if (stricmp(param, "OFF") == 0) {

        hni->flags &= ~NI_EMAILMEMOS;
        notice(s_NickServ, source, "Your memos will no longer be emailed.");
        show_next_db(source, s_NickServ);

    } else {

        notice(s_NickServ, source, "Syntax: \2SET EMAILMEMOS\2 [\37ON\37|\37OFF\37]");
        notice(s_NickServ, source,
                "\2/msg %s HELP SET EMAILMEMOS\2 for more information.",
                s_NickServ);
    }
}


/*************************************************************************/

static void do_set_neverop(NickInfo *ni, char *param)
{
    NickInfo *hni = host(ni);
   char *source = ni->nick;

   if (stricmp(param, "ON") == 0) {
       notice(s_NickServ, source, "\2Never Op\2 setting enabled");
       hni->flags |= NI_NEVEROP;
       show_next_db(source, s_NickServ);
       return;

   } else if (stricmp(param, "OFF") == 0) {
       notice(s_NickServ, source, "\2Never Op\2 setting disabled");
       hni->flags &= ~NI_NEVEROP;
       show_next_db(source, s_NickServ);
       return;

   } else {

       notice(s_NickServ, source, "Syntax: \2SET NEVEROP\2 [\37ON\37|\37OFF\37]");
       notice(s_NickServ, source,
              "\2/msg %s HELP SET NEVEROP\2 for more information.",
              s_NickServ);

  }
}



/*************************************************************************/

static void do_set_noop(NickInfo *ni, char *param)
{
   NickInfo *hni = host(ni);
   char *source = ni->nick;

   if (stricmp(param, "ON") == 0) {
       notice(s_NickServ, source, "\2No Op\2 setting enabled");
       hni->flags |= NI_NOOP;
       show_next_db(source, s_NickServ);
       return;

   } else if (stricmp(param, "OFF") == 0) {
       notice(s_NickServ, source, "\2No Op\2 setting disabled");
       hni->flags &= ~NI_NOOP;
       show_next_db(source, s_NickServ);
       return;

   } else {

       notice(s_NickServ, source, "Syntax: \2SET NOOP\2 [\37ON\37|\37OFF\37]");
       notice(s_NickServ, source,
              "\2/msg %s HELP SET NOOP\2 for more information.",
              s_NickServ);

  }
}

/*************************************************************************/

static void do_set_nomemo(NickInfo *ni, char *param)
{
   NickInfo *hni = host(ni);
   char *source = ni->nick;

   if (stricmp(param, "ON") == 0) {
       notice(s_NickServ, source, "\2No Memo\2 setting enabled");
       hni->flags |= NI_NOMEMO;
       show_next_db(source, s_NickServ);
       return;

   } else if (stricmp(param, "OFF") == 0) {
       notice(s_NickServ, source, "\2No Memo\2 setting disabled");
       hni->flags &= ~NI_NOMEMO;
       show_next_db(source, s_NickServ);
       return;

   } else {

       notice(s_NickServ, source, "Syntax: \2SET NOMEMO\2 [\37ON\37|\37OFF\37]");
       notice(s_NickServ, source,
              "\2/msg %s HELP SET NOMEMO\2 for more information.",
              s_NickServ);

  }
}

static void do_set_nohost(NickInfo *ni, char *param)
{
   NickInfo *hni = host(ni);
   char *source = ni->nick;

   if (stricmp(param, "ON") == 0) {
       notice(s_NickServ, source, "\2No Host Display\2 setting enabled");
       hni->flags |= NI_NOHOST;
       show_next_db(source, s_NickServ);
       return;

   } else if (stricmp(param, "OFF") == 0) {
       notice(s_NickServ, source, "\2No Host Display\2 setting disabled");
       hni->flags &= ~NI_NOHOST;
       show_next_db(source, s_NickServ);
       return;

   } else {

       notice(s_NickServ, source, "Syntax: \2SET NOHOST\2 [\37ON\37|\37OFF\37]");
       notice(s_NickServ, source,
              "\2/msg %s HELP SET NOHOST\2 for more information.",
              s_NickServ);
   }
}

/*************************************************************************/
static void do_link(const char *source)
{
    NickInfo *ni, *hni;
    User *u;
    char *hostnick = strtok (NULL, " ");
    char *pass = strtok (NULL, " ");
    if (!pass)
    {
        notice (s_NickServ, source, "Syntax: \2LINK Nick Password\2");
        notice (s_NickServ, source, "No password set");
    }
    else if (!(u = finduser (source)))
    {
        notice (s_NickServ, source, "Where are you?");
    }
    else if (ni = findnick (source))
        if (ni->flags & NI_VERBOTEN)
            notice (s_NickServ, source, "This nick is forbided");
        else
            notice (s_NickServ, source, "This nick was taken");
    else if (!(hni = findnick (hostnick)))
        notice (s_NickServ, source, "You master nick is not registered");
    else if (strcmp (hni->pass, pass) != 0)
        notice (s_NickServ, source, "Wrong password");
    else if (ni = makenick (source)) {
        strscpy (ni->pass, "", PASSMAX);
        ni->email = NULL;
        ni->url = NULL;
        if (hni->last_usermask)
            free (hni->last_usermask);
        hni->last_usermask = smalloc (strlen (u->nick) + strlen (u->username) + strlen (u->host) + 3);
        snprintf (hni->last_usermask, strlen (u->nick) + strlen (u->username) + strlen (u->host) + 3,
                                "%s@%s", u->username, u->host);
        ni->last_usermask = sstrdup (hni->nick);
        ni->last_realname = sstrdup (u->realname);
        ni->time_registered = ni->last_seen = hni->last_seen = time(NULL);
        ni->flags = NI_IDENTIFIED | NI_RECOGNIZED | NI_SLAVE;
        notice (s_NickServ, source, "You are nick %s linked with %s", source, hni->nick);

    }

}

static void do_access(const char *source)
{
    char *cmd = strtok(NULL, " ");
    char *mask = strtok(NULL, " ");
    NickInfo *ni;
    User *u;
    int i;
    char **access;

    if (cmd && stricmp(cmd, "LIST") == 0 && mask && is_services_admin(source)) {

	if (!(ni = findnick(mask))) {
	    notice(s_NickServ, source, "Nick \2%s\2 not registered!", mask);
	    return;
	}
	notice(s_NickServ, source, "Access list for \2%s\2:", mask);
	for (access = ni->access, i = 0; i < ni->accesscount; ++access, ++i)
	    notice(s_NickServ, source, "    %s", *access);

    } else if (!cmd || (((stricmp(cmd,"LIST")==0) || (stricmp(cmd,"WIPE")==0)) ? !!mask : !mask)) {

	notice(s_NickServ, source,
                "Syntax: \2ACCESS\2 [ADD|DEL|LIST|WIPE] [mask]");
	notice(s_NickServ, source,
		"\2/msg %s HELP ACCESS\2 for more information.", s_NickServ);

    } else if (mask && ((!strchr(mask, '@') || strchr(mask, '!')) &&
               (stricmp(cmd, "ADD") == 0))) {

	notice(s_NickServ, source,
		"Mask must be in the form \37user\37@\37host\37.");
	notice(s_NickServ, source,
		"\2/msg %s HELP ACCESS\2 for more information.", s_NickServ);

    } else if (!(ni = findnick(source))) {

	notice(s_NickServ, source, "Your nick isn't registered.");

    } else if (!(u = finduser(source)) || !(ni->flags & NI_IDENTIFIED)) {

	notice(s_NickServ, source,
		"Password authentication required for that command.");
	notice(s_NickServ, source,
		"Retry after typing \2/msg %s IDENTIFY \37password\37.",
		s_NickServ);
	if (!u)
	    log("%s: SET from nonexistent user %s", s_NickServ, source);

    } else if (stricmp(cmd, "ADD") == 0) {

       if (strlen(mask) > 60) {
           notice(s_NickServ, source, "NickServ access mask too large!");
           wallops(s_NickServ, "%s tried to add a HUGE mask to the access list", source);
           return;
       }
       if (stricmp(mask, "*@*") == 0 || stricmp(mask, "*@*.*") == 0) {
           notice(s_NickServ, source,
               "Please try to add a more secure access mask");
           return;
        }

	for (access = ni->access, i = 0; i < ni->accesscount; ++access, ++i) {
	    if (strcmp(*access, mask) == 0) {
		notice(s_NickServ, source,
			"Mask \2%s\2 already present on your access list.",
			*access);
		return;
	    }
	}
        if ((ni->accesscount > USER_ACCESS_MAX-1) && (USER_ACCESS_MAX > 0)) {
            notice(s_NickServ, source,
                 "Sorry, access list full for nick %s", ni->nick);
            wallops(SERVER_NAME, "%s tried to add another access mask", ni->nick);
            return;
        }
            
	++ni->accesscount;
	ni->access = srealloc(ni->access, sizeof(char *) * ni->accesscount);
	ni->access[ni->accesscount-1] = sstrdup(mask);
	notice(s_NickServ, source, "\2%s\2 added to your access list.", mask);
        show_next_db(source, s_NickServ);

        if (readonly)
            notice(s_NickServ, source,
               "Warning: Services is in read-only mode; changes will not be saved.");

    } else if (stricmp(cmd, "DEL") == 0) {
        int z, j=0;

        if (strspn(mask, "1234567890") == strlen(mask) &&
                            (i = atoi(mask)) > 0 && i <= ni->accesscount) {

            for (access = ni->access, z = 0; z < ni->accesscount;
                        ++access, ++z) {

               j += 1;
               if (i == j) {
                   access = &ni->access[z];
                   break;
               }
               
            }
            if (!(i == j)) {
                notice(s_NickServ, source,
                        "No such entry (#%d) on %s access list.", i, mask);
                return;
            }
            i = z;

        } else {

           for (access = ni->access, i = 0; i < ni->accesscount; ++access, ++i) {
               if (strcmp(*access, mask) == 0)
                    break;
           }
           if (i == ni->accesscount) {
               for (access = ni->access, i = 0; i < ni->accesscount;
                                                     ++access, ++i) {
               if (stricmp(*access, mask) == 0)
		    break;
               }
           }
           if (i == ni->accesscount) {
               notice(s_NickServ, source,
			"\2%s\2 not found on your access list.", mask);
               return;
           }
        }

        notice(s_NickServ, source,
             "\2%s\2 deleted from your access list.", *access);
        show_next_db(source, s_NickServ);
        if (readonly)
            notice(s_NickServ, source,
               "Warning: Services is in read-only mode; changes will not be saved.");

	free(*access);
	--ni->accesscount;
	if (i < ni->accesscount)	/* if it wasn't the last entry... */
	    bcopy(access+1, access, (ni->accesscount-i) * sizeof(char *));
	if (ni->accesscount)		/* if there are any entries left... */
	    ni->access = srealloc(ni->access, ni->accesscount * sizeof(char *));
	else {
	    free(ni->access);
	    ni->access = NULL;
	}

    } else if (stricmp(cmd, "LIST") == 0) {

	notice(s_NickServ, source, "\2Access list:\2");
	for (access = ni->access, i = 0; i < ni->accesscount; ++access, ++i) {
	    if (mask && !match_wild(mask, *access))
		continue;
            notice(s_NickServ, source, " %d)  %s", i+1, *access);
	}
	notice(s_NickServ, source, ERR_EOL);

    } else if (stricmp(cmd, "WIPE") == 0) {
        for (access = ni->access, i = 0; i < ni->accesscount; ++access, ++i) {
	      free(*access);
	}
        ni->accesscount=0;
        free(ni->access);
        ni->access = NULL;
	notice(s_NickServ,source,"Your access list has been wiped.");

    } else {

	notice(s_NickServ, source,
                "Syntax: \2ACCESS\2 [ADD|DEL|LIST|WIPE] [mask]");
	notice(s_NickServ, source,
		"\2/msg %s HELP ACCESS\2 for more information.", s_NickServ);

    }
}

/*************************************************************************/

static void do_acc(const char *source)
{
 char *nick = strtok(NULL, " ");
 NickInfo *ni;

 if (!nick) {
     notice(s_NickServ, source, "Syntax: \2ACC\2 <nick>");
     notice(s_NickServ, source,
         "\2/msg %s HELP ACC\2 for more information.", s_NickServ);
          return;

    } else if ((ni = findnick(nick)) && (ni->flags & NI_VERBOTEN)) {
	notice(s_NickServ, source,
                "Nick \2%s\2: 5 [FORBIDDEN]", nick);
           return;

    } else if (!(ni = findnick(nick))) {
        notice(s_NickServ, source, "Nick \2%s\2: 0 [UnRegistered]", nick);
          return;

    } else if (!finduser(nick)) {
        notice(s_NickServ, source, "Nick \2%s\2: 1 [OffLine]", nick);
         return;

    } else if ((ni = findnick(nick)) && (ni->flags & NI_IDENTIFIED)) {
         notice(s_NickServ, source,
                 "Nick \2%s\2: 3 [Identified]", nick); 
           return;

    } else if ((ni = findnick(nick)) && (ni->flags & NI_RECOGNIZED)) {
        notice(s_NickServ, source, "\2%s\2: 4 Recognized [Access List]", nick);
           return;

    } else if ((ni = findnick(nick)) && !(ni->flags & NI_IDENTIFIED)) {
	notice(s_NickServ, source,
                "Nick \2%s\2: 2 [UnIdentified]", nick);
           return;


  } else {

     notice(s_NickServ, source, "Nick \2%s\2: Unable to Retreive Data", nick);
   }
}



/*************************************************************************/
static void do_info(const char *source)
{
    char *nick = strtok(NULL, " ");
    char *tmp, timebuf[64];
    NickInfo *ni, *hni;
    time_t tbuff;
    User *u;

    if (!nick) {
        notice(s_NickServ, source, "Syntax: \2INFO\2 <nick>");
	notice(s_NickServ, source,
		"\2/msg %s HELP IDENTIFY\2 for more information.", s_NickServ);

    } else if (is_servbot(nick)) {
	do_infoservices(source, nick);

    } else if (!(ni = findnick(nick)) || !(hni = host(ni))) {
	notice(s_NickServ, source, "Nick \2%s\2 isn't registered.", nick);

    } else if (ni->flags & NI_VERBOTEN) {
	notice(s_NickServ, source,
		"Nick \2%s\2 may not be registered or used.", nick);

    } else {
	struct tm tm;
	char buf[512];
	int i = 0;
        u = finduser(nick);

        if (u && (ni->flags & NI_IDENTIFIED)) {
            notice(s_NickServ, source,
               "(\2Currently on IRC\2 [Identified]) - \"/whois %s\" for more information", nick);
	    if (is_services_admin(nick) && is_oper(nick))
            notice(s_NickServ, nick, "%s do NS INFO on you.", source);
	    i = 1;
	}
        else if (u && (ni->flags & NI_RECOGNIZED)) {
            notice(s_NickServ, source,
               "(\2Currently on IRC\2 [Recognized]) - \"/whois %s\" for more information", nick);
	    if (is_services_admin(nick) && is_oper(nick))
            notice(s_NickServ, nick, "%s do NS INFO on you.", source);
	    i = 1;
	}

        if (hni->last_realname)
            notice(s_NickServ, source,
	    	"\2%s\2 is \"%s\"\n", nick, hni->last_realname);
	notice(s_NickServ, source,
		"           Access: %s\n", get_os_access(hni));
//	if ((!(hni->flags & NI_NOHOST) && hni->last_usermask) || (is_services_admin(source) && hni->last_usermask))
	    notice(s_NickServ, source,
		    "Last seen address: %s\n", hni->last_usermask);

	tm = *localtime(&hni->time_registered);
        strftime(timebuf, sizeof(timebuf), "%b %d %H:%M:%S %Y %Z", &tm);
        timebuf[sizeof(timebuf)-1] = 0;
        notice(s_NickServ, source,
                "  Time registered: %s", timebuf);
        if (i == 0) {
	tm = *localtime(&hni->last_seen);
        strftime(timebuf, sizeof(timebuf), "%b %d %H:%M:%S %Y %Z", &tm);
        timebuf[sizeof(timebuf)-1] = 0;
	notice(s_NickServ, source,
		"   Last seen time: %s", timebuf);
        tbuff = time(NULL);
        tm = *localtime(&tbuff);
        strftime(timebuf, sizeof(timebuf), "%b %d %H:%M:%S %Y %Z", &tm);
        timebuf[sizeof(timebuf)-1] = 0;
	notice(s_NickServ, source,
		"         Time Now: %s", timebuf);
	}

        if ((hni->email) && !(hni->flags & NI_HIDE_EMAIL))
	    notice(s_NickServ, source, "   E-mail address: %s\n", hni->email);
	else
	  if (is_services_helpop(source))
	    notice(s_NickServ, source, "   E-mail address: %s\n", hni->email);

	*buf = 0;
        if (hni->flags & NI_HOP)
	    strcat(buf, "IRCop");
	if (hni->flags & NI_KILLPROTECT) {
	    if (*buf)
		strcat(buf, ", ");
	    strcat(buf, "Kill protection");
	}
	if (hni->flags & NI_SECURE) {
	    if (*buf)
		strcat(buf, ", ");
	    strcat(buf, "Security");
	}
        if (hni->flags & NI_NEVEROP) {
            if (*buf)
                strcat(buf, ", ");
            strcat(buf, "NEVER OP");
        }
        if (hni->flags & NI_NOOP) {
            if (*buf)
                strcat(buf, ", ");
            strcat(buf, "NO OP");
        }
        if (hni->flags & NI_NOMEMO) {
            if (*buf)
                strcat(buf, ", ");
            strcat(buf, "NO MEMO");
        }
/*
        if (hni->flags & NI_NOHOST) {
            if (*buf)
                strcat(buf, ", ");
            strcat(buf, "NO HOST");
        }
*/
        if (hni->flags & NI_HIDE_EMAIL) {
            if (*buf)
                strcat(buf, ", ");
            strcat(buf, "Private");
        }

	if (!*buf)
	    strcpy(buf, "None");
	notice(s_NickServ, source, "          Options: %s", buf);

        if (hni->flags & NI_AUTH) {
           notice(s_NickServ, source,
               " * Nick is protected by Authorized Services.");
        }
	if (hni->flags & NI_EXMAIL) {
           notice(s_NickServ, source,
               " * Nick is nearly expired.");
	}
        if (hni->forward)
           notice(s_NickServ, source,
               " * Nick is forwarding memos to: %s", hni->forward);

        if (hni->flags & NI_SOP)
           notice(s_NickServ, source,
               " * Nick is Services Operator");

        if (is_oper(source)) {
            if (hni->flags & NI_MARK)
                notice(s_NickServ, source,
                   " * Nick marked by: %s", hni->mark);
            if (hni->flags & NI_HOLD)
                notice(s_NickServ, source,
                   " * Nick Held by  : %s", hni->hold);
           if (hni->flags & NI_FREEZE)
               notice(s_NickServ, source,
                  " * This nick has been frozed");
	    if (hni->mlock)
               notice(s_NickServ, source,
                   " * Mode Lock     : %s", hni->mlock);
        } else {
           if (hni->flags & NI_MARK)
               notice(s_NickServ, source,
                  " * This nick has been marked");
           if (hni->flags & NI_HOLD)
               notice(s_NickServ, source,
                  " * This nick has been held");
           if (hni->flags & NI_FREEZE)
               notice(s_NickServ, source,
                  " * This nick has been frozed");
       }                       

	if (hni->age)
               notice(s_NickServ, source,
                  " * \37FIELDS\37: \2AGE\2   : %s years old", hni->age);
	if (hni->sex)
               notice(s_NickServ, source,
                  " * \37FIELDS\37: \2SEX\2   : %s", hni->sex);
        if (hni->uin)  
               notice(s_NickServ, source,
		  " * \37FIELDS\37: \2UIN\2   : %s\n",hni->uin); 
	if (hni->url)
	       notice(s_NickServ, source,
		  " * \37FIELDS\37: \2URL\2   : %s\n", hni->url);
	if (hni->info)
               notice(s_NickServ, source,
                  " * \37FIELDS\37: \2INFO\2  : %s", hni->info); 
/*
	if (ni->auth && is_services_admin(source))
               notice(s_NickServ, source,
                  " * \37FIELDS\37: \2AUTH\2  : %s", hni->auth); 
*/
    }
}

/*************************************************************************/

static void do_list(const char *source)
{
    char *pattern = strtok(NULL, " ");
    NickInfo *ni;
    int nnicks, i;
    char buf[BUFSIZE];
    int is_servadmin = is_services_admin(source);

    if (!pattern) {

        notice(s_NickServ, source, "Syntax: \2LIST\2 [pattern]");
	notice(s_NickServ, source,
		"\2/msg %s HELP LIST\2 for more information.", s_NickServ);

    } else {

	nnicks = 0;
	notice(s_NickServ, source, "List of entries matching \2%s\2:",
		pattern);
	for (i = 33; i < 256; ++i) {
	    for (ni = nicklists[i]; ni; ni = ni->next) {
		if (ni->flags & NI_VERBOTEN)
		    continue;
		if (!is_servadmin && (ni->flags & NI_PRIVATE))
		    continue;
		snprintf(buf, sizeof(buf), "%-20s  %s", ni->nick, ni->last_usermask);
		if (stricmp(pattern, ni->nick) == 0 ||
					match_wild_nocase(pattern, buf)) {
		    if (++nnicks <= 50)
			notice(s_NickServ, source, "    %s", buf);
		}
	    }
	}
	notice(s_NickServ, source, "End of list - %d/%d matches shown.",
					nnicks>50 ? 50 : nnicks, nnicks);
    }

}

/*************************************************************************/

static void do_recover(const char *source)
{
    char *nick = strtok(NULL, " ");
    char *pass = strtok(NULL, " ");
    NickInfo *ni;
    User *u = finduser(source);

    if (!nick) {

	notice(s_NickServ, source,
                "Syntax: \2RECOVER\2 <nickname> <password>");
	notice(s_NickServ, source,
		"\2/msg %s HELP RECOVER\2 for more information.", s_NickServ);

    } else if (!(ni = findnick(nick))) {

	notice(s_NickServ, source, "Nick %s isn't registered.", nick);

    } else if (!finduser(nick)) {

	notice(s_NickServ, source, "Nick %s isn't currently in use.");

    } else if (stricmp(nick, source) == 0) {

	notice(s_NickServ, source, "You can't recover yourself.");

    } else if (u &&
	    (pass ? check_password(pass,ni->pass)==1 :
	     !(ni->flags & NI_SECURE) && is_on_access(u,ni))) {

	collide(ni, 0);
	notice(s_NickServ, source, "User claiming your nick has been killed.");
	notice(s_NickServ, source,
		"\2/msg %s RELEASE\2 to get it back before the one-minute "
		"timeout.", s_NickServ);

    } else {

	notice(s_NickServ, source, "Access denied.");
	if (!u)
	    log("%s: RECOVER: source user %s not found!", s_NickServ, source);
	else if (pass) {
	    log("%s: RECOVER: invalid password for %s by %s!%s@%s",
                        s_NickServ, nick, source, u->username, u->host);
           if (EXTRA_SNOOP) {
                slog("NS *R %s (%s!%@%s)",
                     nick, source, u->username, u->host);
                 do_break_log("NS_O", "NS *R %s (%s!%@%s)",
                     nick, source, u->username, u->host);
           }
	   bad_password(u, nick);
	}

    }
}

/*************************************************************************/

static void do_release(const char *source)
{
    char *nick = strtok(NULL, " ");
    char *pass = strtok(NULL, " ");
    NickInfo *ni;
    User *u = finduser(source);
    int i;

    if (!nick) {

	notice(s_NickServ, source,
                "Syntax: \2RELEASE\2 <nickname> <password>");
	notice(s_NickServ, source,
		"\2/msg %s HELP RELEASE\2 for more information.", s_NickServ);

    } else if (!(ni = findnick(nick))) {

	notice(s_NickServ, source, "Nick %s isn't registered.", nick);

    } else if (!(ni->flags & NI_KILL_HELD)) {

	notice(s_NickServ, source, "Nick %s isn't being held.", nick);

    } else if (u &&
	    (pass ? check_password(pass,ni->pass)==1 :
	     !(ni->flags & NI_SECURE) && is_on_access(u,ni))) {

	release(ni, 0);
	notice(s_NickServ, source,
		"Services' hold on your nick has been released.");

        if (!is_on_id_list(u->nick, nick)) {
            for (i=0; u->id_nicks[i] && i < MAX_IDS; i++)
                ;
            if (i < MAX_IDS)
                 u->id_nicks[i] = sstrdup(nick);
        }

    } else {

	notice(s_NickServ, source, "Access denied.");
	if (!u)
	    log("%s: RELEASE: source user %s not found!", s_NickServ, source);
	else if (pass) {
	    log("%s: RELEASE: invalid password for %s by %s!%s@%s",
                        s_NickServ, nick, source, u->username, u->host);
            if (EXTRA_SNOOP) {
                slog("NS *Re %s (%s!%s@%s)",
                    nick, source, u->username, u->host);
                do_break_log("NS_O", "NS *Re %s (%s!%s@%s)",
                    nick, source, u->username, u->host);
            }
	    bad_password(u, nick);
	}

    }
}

/*************************************************************************/
static void do_ghost(const char *source)
{
    char *nick = strtok(NULL, " ");
    char *pass = strtok(NULL, " ");
    NickInfo *ni, *hni;
    User *u = finduser(source);
    int i;

    if (!nick) {

	notice(s_NickServ, source,
                "Syntax: \2GHOST\2 <nickname> <password>");
	notice(s_NickServ, source,
		"\2/msg %s HELP GHOST\2 for more information.", s_NickServ);

    } else if (!(ni = findnick(nick)) || !(hni = host(ni))) {

	notice(s_NickServ, source, "Nick %s isn't registered.", nick);

    } else if (!finduser(nick)) {

	notice(s_NickServ, source, "Nick %s isn't currently in use.", nick);

    } else if (stricmp(nick, source) == 0) {

	notice(s_NickServ, source, "You can't ghost yourself!");

    } else if (u &&
	    (pass ? check_password(pass,hni->pass)==1 :
	     !(ni->flags & NI_SECURE) && is_on_access(u,ni))) {

	char buf[NICKMAX+32];

	snprintf(buf, sizeof(buf), "Ghost command used by %s", source);
        kill_user(SERVER_NAME, nick, buf);
	notice(s_NickServ, source, "Ghost with your nick has been killed.");

        if (!is_on_id_list(u->nick, hni->nick)) {
            for (i=0; u->id_nicks[i] && i < MAX_IDS; i++)
                ;
            if (i < MAX_IDS)
                 u->id_nicks[i] = sstrdup(hni->nick);
        }


    } else {

	notice(s_NickServ, source, "Access denied.");
	if (!u)
	    log("%s: GHOST: source user %s not found!", s_NickServ, source);
	else if (pass) {
	    log("%s: GHOST: invalid password for %s by %s!%s@%s",
                        s_NickServ, nick, source, u->username, u->host);
            if (EXTRA_SNOOP) {
                slog("NS *K %s (%s!%s@%s)",
                     nick, source, u->username, u->host);
                do_break_log("NS_O", "NS *K %s (%s!%s@%s)",
                     nick, source, u->username, u->host);
            }
	    bad_password(u, nick);
	}

    }
}

/*************************************************************************/

static void do_sendpass(const char *source)
{
    char *nick = strtok(NULL, " ");
    NickInfo *ni;
    User *u = finduser(source);

    if (!nick) {
        notice(s_NickServ, source, "Syntax: \2SENDPASS\2 <nickname>");
    } else if (!u) {
	notice(s_NickServ, source, "Couldn't get your user info!");
    } else if (!(ni = findnick(nick))) {
	notice(s_NickServ, source, NS_NOTREG, nick);
    } else if (ni->flags & NI_SLAVE) {
        notice(s_NickServ,source,"Nick \2%s\2 is a slave nick!",nick);
    } else if (ni->flags & NI_MARK) {
        slog("NS *S %s (%s!%s@%s)",
            nick, u->nick, u->username, u->host);
        do_break_log("NS_G", "NS *S %s (%s!%s@%s)",
            nick, u->nick, u->username, u->host);
            notice(s_NickServ,source,"Nick \2%s\2 is marked!",nick);
        log("%s: %s!%s@%s attempted SENDPASS on %s (Fail)"
                ,s_NickServ, source, u->username, u->host, nick);

    } else if (!ni->email) {
          notice(s_NickServ, source,
               "User %s doesn't have an E-Mail specified", ni->nick);

    } else {
	log("%s: %s!%s@%s used SENDPASS on %s",
		s_NickServ, source, u->username, u->host, nick);
        slog("NS S %s (%s!%s@%s)",
            nick, u->nick, u->username, u->host);
        slog("NS S %s (%s!%s@%s)",
            nick, u->nick, u->username, u->host);
        wallops(s_NickServ, "%s used SENDPASS on %s", source, nick);
        notice(s_NickServ, source,
              "Password sent to %s", ni->email);
        sendpass_ns(source, ni);
    }
}


/*************************************************************************/

static void do_getpass(const char *source)
{
    char *nick = strtok(NULL, " ");
    NickInfo *ni;
    User *u = finduser(source);

    if (!nick) {
        notice(s_NickServ, source, "Syntax: \2GETPASS\2 <nickname>");
    } else if (!u) {
	notice(s_NickServ, source, "Couldn't get your user info!");
    } else if (!(ni = findnick(nick))) {
	notice(s_NickServ, source, "Nick %s isn't registered.", nick);
    } else if (is_services_admin(nick) && !is_services_coder(source)
          && !stricmp(source, nick) == 0) {
        notice(s_NickServ, source,
           "You may not getpass a services administrator");

    } else if (ni->flags & NI_MARK) {
        slog("NS *G %s (%s!%s@%s)",
            nick, u->nick, u->username, u->host);
        do_break_log("NS_G", "NS *G %s (%s!%s@%s)",
            nick, u->nick, u->username, u->host);
            notice(s_NickServ,source,"Nick \2%s\2 is marked!",nick);
        log("%s: %s!%s@%s attempted GETPASS on %s (Fail)"
                ,s_NickServ, source, u->username, u->host, nick);
    } else {
	log("%s: %s!%s@%s used GETPASS on %s",
		s_NickServ, source, u->username, u->host, nick);
        slog("NS G %s (%s!%s@%s)",
            nick, u->nick, u->username, u->host);
        slog("NS G %s (%s!%s@%s)",
            nick, u->nick, u->username, u->host);
        wallops(s_NickServ, "%s used GETPASS on %s", source, nick);
	notice(s_NickServ, source, "Password for %s is \2%s\2.",
		nick, ni->pass);
    }
}

/*************************************************************************/

static void do_forbid(const char *source)
{
    NickInfo *ni;
    User *u = finduser(source);
    char *nick = strtok(NULL, " ");

    if (!is_services_oper(source)) {
	notice(s_NickServ,source,"Permission Denied.");
	return;
    }

    if (!nick) {
        notice(s_NickServ, source, "Syntax: \2FORBID\2 [-]<nickname>");
        notice(s_NickServ,source,
             "\2/msg %s OHELP FORBID\2 for more information",s_NickServ);
	return;
    }
    if (*nick == '-') {
       *nick++;
       do_unforbid(source, nick);
       return;
    }
    if (readonly) {
	notice(s_NickServ, source,
	    "Warning: Services is in read-only mode; changes will not be saved.");
    }
    if (is_services_oper(nick)) {
           notice(s_NickServ, source, "Error: Nick is a services admin");
           wallops(s_NickServ, "%s tried to forbid \2%s\2", source, nick);
           return;
    }
    if (ni = findnick(nick)) {
        if (ni->forbid) {
           notice(s_NickServ, source, "Nick %s already forbidden", nick);
           return;
        }
	delnick(ni);
    }
    if (ni = makenick(nick)) {
	ni->flags |= NI_VERBOTEN;
        if (ni->forbid)
             free(ni->forbid);
        ni->forbid = sstrdup(u->nick);
        slog("NS +F %s (%s!%s@%s)",
             nick,u->nick,u->username,u->host);
        do_break_log("NS_F", "NS +F %s (%s!%s@%s)",
             nick,u->nick,u->username,u->host);
	log("%s: %s set FORBID for nick %s", s_NickServ, source, nick);
        wallops(s_NickServ, "%s set \2%s\2 as FORBIDDEN ", source, nick);
	notice(s_NickServ, source, "Nick \2%s\2 is now FORBIDden.", nick);
    } else {
	log("%s: Valid FORBID for %s by %s failed", s_NickServ,
		nick, source);
	notice(s_NickServ, source, "Couldn't FORBID nick \2%s\2!", nick);
    }
}

static void do_unforbid(const char *source, const char *nick)
{
    NickInfo *ni;
    NickInfo *ni1;
    User *u = finduser(source);

    if (!nick) {
        notice(s_NickServ, source, "Syntax: \2FORBID\2 [-]<nick>");
        notice(s_NickServ, source,
            "\2/MSG %s OHELP FORBID\2 for more information", s_NickServ);
	return;
    }
    if (readonly) {
	notice(s_NickServ, source,
	    "Warning: Services is in read-only mode; changes will not be saved.");
    }

   if (ni = findnick(nick)) {
       if (!(ni->flags & NI_VERBOTEN)) {
	 if (((ni->forbid) && (ni1=findnick(ni->forbid)) && (ni1->flags & NI_IDENTIFIED) && (stricmp(source, ni->forbid) == 0)) || is_services_root(source)) {
           slog("NS -F* %s (%s!%s@%s)",
                nick,u->nick,u->username,u->host);
           do_break_log("NS_F", "NS -F* %s (%s!%s@%s)",
                nick,u->nick,u->username,u->host);

           log("%s: %s tried to set unforbidden flag for non forbidden nick %s",
                   s_NickServ, source, nick);
           if (ni->forbid)
               free(ni->forbid);
           ni->forbid = NULL;

           notice(s_NickServ, source, "%s is not set as FORBIDDEN", nick);
           return;
	 } else {
           notice(s_NickServ, source, "This nick is Forbidden by other admin.");
           return;
	 }
        }
	if (((ni->forbid) && (ni1=findnick(ni->forbid)) && (ni1->flags & NI_IDENTIFIED) && (stricmp(source, ni->forbid) == 0)) || is_services_root(source)) {
        slog("NS -F %s (%s!%s@%s) [Dropped]",
             nick,u->nick,u->username,u->host);
        do_break_log("NS_F", "NS -F %s (%s!%s@%s) [Dropped]",
             nick,u->nick,u->username,u->host);
        log("%s: %s set FORBID for nick %s", s_NickServ, source, nick);
        delnick(ni);
        wallops(s_NickServ, "%s removed \2%s\2 from the FORBIDDEN list [Automatically Dropped]", source, nick);
        notice(s_NickServ, source, "Nick %s is now UNFORBIDDEN.", nick);
	} else
           notice(s_NickServ, source, "This nick is Forbidden by other admin.");
    } else {
        log("%s: Valid UNFORBID for %s by %s failed", s_NickServ,
		nick, source);
        notice(s_NickServ, source, "Couldn't UNFORBID nick \2%s\2!", nick);
    }
}

static char *get_os_access(NickInfo *ni)
{
    char *access, buf[128], sbuf[128];
    int i=0, end=0;

    snprintf(buf, sizeof(buf), "Services User");
   
    for (i = 0; i < MAX_SERVROOTS; i++) {
        if (services_roots[i]) {
           if (stricmp(services_roots[i], ni->nick) == 0) {
              end++;
              if (stricmp(services_roots[i], SERVICES_MASTER) == 0)
                   snprintf(buf, sizeof(buf), "Services Root [Master]");
	      else if (stricmp(services_roots[i], SERVICES_MASTER2) == 0)
                   snprintf(buf, sizeof(buf), "Services Root [Master]");
	      else if (stricmp(services_roots[i], SERVICES_MASTER3) == 0)
                   snprintf(buf, sizeof(buf), "Services Root [Master]");
              else
                   snprintf(buf, sizeof(buf), "Services Root");
           }
        }
    }

    for (i = 0; i < MAX_SERVADMINS; i++) {
        if (services_admins[i]) {
           if (stricmp(services_admins[i], ni->nick) == 0) {
                   if (!end) {
                      snprintf(buf, sizeof(buf), "Services Admin");
                      end++;
                   }
           }
        }
    }
   if (ni->flags & NI_SOP)
      if (!end) {
      	snprintf(buf, sizeof(buf), "Services Operator");
        end++;
      }

   if (ni->flags & NI_HOP)
      if (!end) {
        if (is_oper(ni->nick))
	        snprintf(buf, sizeof(buf), " ");
	else
        	snprintf(buf, sizeof(buf), "Help Operator");
        end++;
      }

    if (is_netadmin(ni->nick)) {
         snprintf(sbuf, sizeof(sbuf), "%s, a Network Administrator", buf);
         end++;
    } else if (is_techadmin(ni->nick)) {
         snprintf(sbuf, sizeof(sbuf), "%s, a Technical Administrator", buf);
         end++;
    } else if (is_servadmin(ni->nick)) {
         snprintf(sbuf, sizeof(sbuf), "%s, a Server Administrator", buf);
         end++;
    } else if (is_oper(ni->nick)) {
         snprintf(sbuf, sizeof(sbuf), "%s, an IRC Operator", buf);
         end++;
    } else {
         snprintf(sbuf, sizeof(sbuf), "%s", buf);
    }

    if (!end)
       snprintf(sbuf, sizeof(sbuf), "User");

    access = sbuf;
    return access;
}

static void do_flist(const char *source)
{
    char *cmd = strtok(NULL, " ");
    NickInfo *ni;
    int i;
    int totalaccess = 0;

    if (!is_services_oper(source)) {
	notice(s_NickServ, source,
		"Unrecognized command \2%s\2.  Type \"/msg %s HELP\" for help.",
		cmd, s_NickServ);
        return;
    }

    if (!cmd) {
        notice(s_NickServ,source,"Syntax: \2FLIST\2 [HELD|MARKED|FORBIDDEN]");
        notice(s_NickServ,source,
            "\2/msg %s OHELP FLIST\2 for more information",s_NickServ);

    } else if (stricmp(cmd, "HELD") == 0) {
        notice(s_NickServ,source,"\2HELD\2 nickname listing:");
        for  (i = 0; i < 256; ++i) {
          for (ni = nicklists[i]; ni; ni = ni->next) {
              if (ni->hold) {
                notice(s_NickServ,source,"%-20s  %s",
                   ni->nick,ni->last_usermask);
                notice(s_NickServ,source,
                     "Set By: %s",ni->hold);
		  ++totalaccess;
              }
          }
        }
        notice(s_NickServ,source, NS_NFEOL, totalaccess);
    } else if (stricmp(cmd, "MARKED") == 0) {
        notice(s_NickServ,source,"\2MARKED\2 nickname listing:");
        for (i = 0; i < 256; ++i) {
            for (ni = nicklists[i]; ni; ni = ni->next) {
                if (ni->mark) {
                  notice(s_NickServ,source,"%-20s  %s",
                        ni->nick,ni->last_usermask);
                notice(s_NickServ,source,
                     "Set By: %s",ni->mark);
		  ++totalaccess;
                }
            }
        }
        notice(s_NickServ,source, NS_NFEOL, totalaccess);
    } else if (stricmp(cmd, "FORBIDDEN") == 0) {
        notice(s_NickServ,source,"\2FORBIDDEN\2 nickname listing:");
        for (i = 0; i < 256; ++i) {
            for (ni = nicklists[i]; ni; ni = ni->next) {
                if (ni->forbid) {
                  notice(s_NickServ,source,"%-20s  %s",
                        ni->nick,ni->last_usermask);
                notice(s_NickServ,source,
                     "Set By: %s",ni->forbid);
		  ++totalaccess;
                }
            }
        }
        notice(s_NickServ,source, NS_NFEOL, totalaccess);
    } else if (stricmp(cmd, "HOP") == 0) {
        notice(s_NickServ,source,"\2HOP\2 nickname listing:");
        for (i = 0; i < 256; ++i) {
            for (ni = nicklists[i]; ni; ni = ni->next) {
                if (ni->flags & NI_HOP) {
                  notice(s_NickServ,source,"%-20s",
                        ni->nick);
		  ++totalaccess;
                }
            }
        }
        notice(s_NickServ,source, NS_NFEOL, totalaccess);
    } else if (stricmp(cmd, "SOP") == 0) {
        notice(s_NickServ,source,"\2SOP\2 nickname listing:");
        for (i = 0; i < 256; ++i) {
            for (ni = nicklists[i]; ni; ni = ni->next) {
                if (ni->flags & NI_SOP) {
                  notice(s_NickServ,source,"%-20s",
                        ni->nick);
		  ++totalaccess;
                }
            }
        }
        notice(s_NickServ,source, NS_NFEOL, totalaccess);
    }
}

/*************************************************************************/

static void do_unhold(const char *source, const char *nick)
{
     User *u = finduser(source);
     NickInfo *ni;
     NickInfo *ni1;
     if (!(ni = findnick(nick))) {
          notice(s_NickServ, source, "Nickname %s is not registered.",nick);
     } else if (ni->flags & NI_HOLD) {
	 if (((ni->hold) && (ni1=findnick(ni->hold)) && (ni1->flags & NI_IDENTIFIED) && (stricmp(source, ni->hold) == 0)) || is_services_root(source)) {
          ni->flags &= ~NI_HOLD;
         if (ni->hold)
             free(ni->hold);
          ni->hold = NULL;
          slog("NS -H %s (%s!%s@%s)",
              nick,u->nick,u->username,u->host);
          do_break_log("NS_F", "NS -H %s (%s!%s@%s)",
               nick,u->nick,u->username,u->host);
          wallops(s_NickServ,
              "%s unset the HELD flag for \2%s\2",u->nick,nick);
	} else
          notice(s_NickServ,source,"This nickname is held by other admin.");
     } else {
          notice(s_NickServ,source,"Nickname \2%s\2 is not held.",nick);
     }
}

/*************************************************************************/

static void do_unmark(const char *source, const char *nick)
{
     User *u = finduser(source);
     NickInfo *ni;
     NickInfo *ni1;

     if (!(ni = findnick(nick))) {
          notice(s_NickServ, source, "Nickname %s is not registered.",nick);
     } else if (ni->flags & NI_MARK) {
	 if (((ni->mark) && (ni1=findnick(ni->mark)) && (ni1->flags & NI_IDENTIFIED) && (stricmp(source, ni->mark) == 0)) || is_services_root(source)) {
          ni->flags &= ~NI_MARK;
         if (ni->mark)
             free(ni->mark);
          ni->mark = NULL;
          slog("NS -M %s (%s!%s@%s)",
              nick,u->nick,u->username,u->host);
          do_break_log("NS_F", "NS -M %s (%s!%s@%s)",
               nick,u->nick,u->username,u->host);
            wallops(s_NickServ,
              "%s unset the MARK flag for \2%s\2",u->nick,nick);
	  } else
          notice(s_NickServ,source,"This nick is marked by other admin.");

     } else {
          notice(s_NickServ,source,"Nickname \2%s\2 is not marked.",nick);
     }

}
/*************************************************************************/
static void do_unfreeze(const char *source, const char *nick)
{
     User *u = finduser(source);
     NickInfo *ni;

     if (!(ni = findnick(nick))) {
          notice(s_NickServ, source, "Nickname %s is not registered.",nick);
     } else if (ni->flags & NI_FREEZE) {
          ni->flags &= ~NI_FREEZE;
          slog("NS -FR %s (%s!%s@%s)",
              nick,u->nick,u->username,u->host);
          do_break_log("NS_F", "NS -FR %s (%s!%s@%s)",
               nick,u->nick,u->username,u->host);
            wallops(s_NickServ,
              "%s unset the FREEZE flag for \2%s\2",u->nick,nick);
     } else {
          notice(s_NickServ,source,"Nickname \2%s\2 is not frozen.",nick);
     }

}
/*************************************************************************/

static void do_setpass(const char *source)
{
    char *nick = strtok (NULL, " ");
    char *newpass = strtok (NULL, " ");
    NickInfo *ni;
    User *u = finduser (source);

    if (!nick || !newpass)
        notice (s_NickServ, source, "Syntax: \2SETPASS Nick Password\2");
    else if (!u)
        notice (s_NickServ, source, "Where are you?");

    else if (!(ni = findnick (nick)))
        notice (s_NickServ, source, "nick %s is not registered. ", nick);

    else if ((ni->flags & NI_MARK) && (!is_services_admin(source)))
        notice (s_NickServ, source, "Can not SETPASS on marked nick. ", nick);
    else
    {
       strscpy (ni->pass, newpass, PASSMAX);
       show_next_db(source, s_NickServ);
       notice(s_NickServ, source, "Password for nick %s changed to \2%s\2.", nick, newpass);
    }
}
/**************************************************************************/
static void do_setmlock(const char *source)
{
    char *nick = strtok (NULL, " ");
    char *modelocks = strtok (NULL, "");
    NickInfo *ni;
    User *u = finduser (source);

    if (!nick || !modelocks)
        notice (s_NickServ, source, "Syntax: \2SETMLOCK umodes\2");
    else if (!u)
        notice (s_NickServ, source, "Where are you?");

    else if (!(ni = findnick (nick)))
        notice (s_NickServ, source, "nick %s is not registered. ", nick);

    else if ((ni->flags & NI_MARK) && (!is_services_admin(source)))
        notice (s_NickServ, source, "Can not SETMLOCK on marked nick. ");
    else
    {
       ni->mlock = sstrdup (modelocks);
       show_next_db(source, s_NickServ);
       notice(s_NickServ, source, "Mlock for nick %s changed to \2%s\2.", nick, modelocks);
    }
}
/**************************************************************************/

static void do_set_hidemail(NickInfo *ni, char *param)
{
    NickInfo *hni;
    char *source = ni->nick;
    hni = host(ni);

    if (stricmp(param,"ON") == 0) {

	 if (hni->flags & NI_HIDE_EMAIL) {
            notice(s_NickServ, ni->nick,
                "Your email already HIDEN");
            return;
	 }

         if (!hni->email) {
            notice(s_NickServ, ni->nick,
                "You have no email to HIDE");
            return;
         }
         hni->flags |= NI_HIDE_EMAIL;
         notice(s_NickServ, ni->nick,
                  "Your email address will no longer be shown");
         show_next_db(source, s_NickServ);

    } else if (stricmp(param,"OFF") == 0) {

       if (!hni->email) {
           notice(s_NickServ, ni->nick,
               "You have no email address set to show!");
           return;
       }
	if (!(hni->flags & NI_HIDE_EMAIL)) {
            notice(s_NickServ, ni->nick,
                "Your email already show");
            return;
	}

       hni->flags &= ~NI_HIDE_EMAIL;
       notice(s_NickServ, source, "Your EMAIL is no longer hide");
       show_next_db(source, s_NickServ);

     } else  {
           notice(s_NickServ, source, "Syntax: SET HIDEMAIL ON/OFF");
     }
}
/************************************************************************/
static void do_set_auth(NickInfo *ni, char *param)
{
   NickInfo *hni = host(ni);
   char *source = ni->nick;

   if (stricmp(param, "ON") == 0) {
       notice(s_NickServ, source, "You need to authorized before add to any channel access");
       hni->flags |= NI_AUTH;
       show_next_db(source, s_NickServ);
       return;

   } else if (stricmp(param, "OFF") == 0) {
       notice(s_NickServ, source, "Disable authorized.");
       hni->flags &= ~NI_AUTH;
       show_next_db(source, s_NickServ);
       return;

   } else {

       notice(s_NickServ, source, "Syntax: \2SET AUTH\2 [\37ON\37|\37OFF\37]");
       notice(s_NickServ, source,
              "\2/msg %s HELP SET AUTH\2 for more information.",
              s_NickServ);

  }
}
/***************************************************************************/
static void do_set_botmsg(NickInfo *ni, char *param)
{
   NickInfo *hni = host(ni);
   char *source = ni->nick;

   if (stricmp(param, "ON") == 0) {
       notice(s_NickServ, source, "BotMsg Enable");
       hni->flags |= NI_BOTMSG;
       show_next_db(source, s_NickServ);
       return;

   } else if (stricmp(param, "OFF") == 0) {
       notice(s_NickServ, source, "BotMsg Disable.");
       hni->flags &= ~NI_BOTMSG;
       show_next_db(source, s_NickServ);
       return;

   } else {

       notice(s_NickServ, source, "Syntax: \2SET BOTMSG\2 [\37ON\37|\37OFF\37]");
       notice(s_NickServ, source,
              "\2/msg %s HELP SET BOTMSG\2 for more information.",
              s_NickServ);

  }
}
/***************************************************************************/
static void do_set_privmsg(NickInfo *ni, char *param)
{
   NickInfo *hni = host(ni);
   char *source = ni->nick;

   if (stricmp(param, "ON") == 0) {
       notice(s_NickServ, source, "You may recieve notice in private message");
       hni->flags |= NI_PRIVMSG;
       show_next_db(source, s_NickServ);
       return;

   } else if (stricmp(param, "OFF") == 0) {
       notice(s_NickServ, source, "You may recieve services message in notice.");
       hni->flags &= ~NI_PRIVMSG;
       show_next_db(source, s_NickServ);
       return;

   } else {

       notice(s_NickServ, source, "Syntax: \2SET PRIVMSG\2 [\37ON\37|\37OFF\37]");
       notice(s_NickServ, source,
              "\2/msg %s HELP SET PRIVMSG\2 for more information.",
              s_NickServ);

  }
}
/**************************************************************************/
static void do_set_malay(NickInfo *ni, char *param)
{
   NickInfo *hni = host(ni);
   char *source = ni->nick;

   if (stricmp(param, "ON") == 0) {
       notice(s_NickServ, source, "Change help language to Malay");
       hni->flags |= NI_MALAY;
       show_next_db(source, s_NickServ);
       return;

   } else if (stricmp(param, "OFF") == 0) {
       notice(s_NickServ, source, "Change help language to English.");
       hni->flags &= ~NI_MALAY;
       show_next_db(source, s_NickServ);
       return;

   } else {

       notice(s_NickServ, source, "Syntax: \2SET MALAY\2 [\37ON\37|\37OFF\37]");
       notice(s_NickServ, source,
              "\2/msg %s HELP SET MALAY\2 for more information.",
              s_NickServ);

  }
}
/******************************************************************/
static void do_setflags(const char *source)
{
    char *nick = strtok (NULL, " ");
    char *nflag = strtok (NULL, " ");
    char *noption = strtok (NULL, " ");
    NickInfo *ni;
    User *u = finduser (source);

    if (!nick || !nflag || !noption)
        notice (s_NickServ, source, "Syntax: \2SETFLAG Nick flag ON/OFF\2");
    else if (!u)
        notice (s_NickServ, source, "Where are you?");

    else if (!(ni = findnick (nick)))
        notice (s_NickServ, source, "nick %s is not registered. ", nick);

    else if ((ni->flags & NI_MARK) && (!is_services_admin(source)))
        notice (s_NickServ, source, "Can not SETFLAG on marked nick. ", nick);

    else if (!(stricmp(noption, "ON")==0) && !(stricmp(noption, "OFF")==0))
        notice (s_NickServ, source, "Option must be ON or OFF");

    else
    {
        if (stricmp(nflag, "AUTH")==0) {
		if(stricmp(noption, "ON")==0)
		       ni->flags |= NI_AUTH;
		else 
		       ni->flags &= ~NI_AUTH;
	       notice(s_NickServ, source, "set flag %s for nick %s \2%s\2.", nflag, nick, noption);

	} else if (stricmp(nflag, "HOP")==0) {
		if (!is_services_admin(source)) {
		        notice(s_NickServ, source, "You need to be Services Admin to set this flag.");
			return;
		}

		if(stricmp(noption, "ON")==0) {
		       ni->flags |= NI_HOP;
		} else 
		       ni->flags &= ~NI_HOP;
	       notice(s_NickServ, source, "set flag %s for nick %s \2%s\2.", nflag, nick, noption);
	
	} else if (stricmp(nflag, "SOP")==0) {

		if (!is_services_admin(source)) {
		        notice(s_NickServ, source, "You need to be Services Admin to set this flag.");
			return;
		}

		if(stricmp(noption, "ON")==0) {
		       ni->flags |= NI_SOP;
		} else 
		       ni->flags &= ~NI_SOP;
	       notice(s_NickServ, source, "set flag %s for nick %s \2%s\2.", nflag, nick, noption);

	} else if (stricmp(nflag, "MALAY")==0) {
		if(stricmp(noption, "ON")==0)
		       ni->flags |= NI_MALAY;
		else 
		       ni->flags &= ~NI_MALAY;
	       notice(s_NickServ, source, "set flag %s for nick %s \2%s\2.", nflag, nick, noption);

	} else if (stricmp(nflag, "PRIVMSG")==0) {
		if(stricmp(noption, "ON")==0)
		       ni->flags |= NI_PRIVMSG;
		else 
		       ni->flags &= ~NI_PRIVMSG;
	       notice(s_NickServ, source, "set flag %s for nick %s \2%s\2.", nflag, nick, noption);

	} else if (stricmp(nflag, "EMAILMEMOS")==0) {

		if (!is_services_admin(source)) {
		        notice(s_NickServ, source, "You need to be Services Admin to set this flag.");
			return;
		}

		if(stricmp(noption, "ON")==0)
		       ni->flags |= NI_EMAILMEMOS;
		else 
		       ni->flags &= ~NI_EMAILMEMOS;
	       notice(s_NickServ, source, "set flag %s for nick %s \2%s\2.", nflag, nick, noption);

	} else if (stricmp(nflag, "NOOP")==0) {
		if(stricmp(noption, "ON")==0)
		       ni->flags |= NI_NOOP;
		else 
		       ni->flags &= ~NI_NOOP;
	       notice(s_NickServ, source, "set flag %s for nick %s \2%s\2.", nflag, nick, noption);

	} else if (stricmp(nflag, "HIDEMAIL")==0) {
		if(stricmp(noption, "ON")==0)
		       ni->flags |= NI_HIDE_EMAIL;
		else 
		       ni->flags &= ~NI_HIDE_EMAIL;
	       notice(s_NickServ, source, "set flag %s for nick %s \2%s\2.", nflag, nick, noption);

	} else if (stricmp(nflag, "KILL")==0) {
		if(stricmp(noption, "ON")==0)
		       ni->flags |= NI_KILLPROTECT;
		else 
		       ni->flags &= ~NI_KILLPROTECT;
	       notice(s_NickServ, source, "set flag %s for nick %s \2%s\2.", nflag, nick, noption);
	}	
    }
}
/***********************************************************************/
void do_set_ossop (const char *source) {
    char *cmd, *nick;
    NickInfo *ni;
    int i;

    cmd = strtok(NULL, " ");
    if (!cmd)
        cmd = "";
    if (stricmp(cmd, "ADD") == 0) {
        if (!is_services_admin(source)) {
            notice(s_OperServ, source, "Permission denied.");
            return;
        }
        nick = strtok(NULL, " ");
        if (nick) {
	    ni = findnick(nick);
            if (!ni) {
                notice(s_OperServ, source, "Nick %s isn't registered!", nick);
                return;
            }
	        ni->flags |= NI_SOP;
		slog("!! Added \2%s\2 to SOP List", nick);
		do_break_log("OS", "!! Added %s to SOP List", nick);
		wallops(SERVER_NAME, "%s added %s to the SOP List", source, nick);
		notice(s_OperServ, source, "%s added to Services Operator list.", nick);
	}
    } else if (stricmp(cmd, "DEL") == 0) {
        if (!is_services_admin(source)) {
            notice(s_OperServ, source, "Permission denied.");
            return;
        }
        nick = strtok(NULL, " ");
        if (nick) {
	    ni = findnick(nick);
            if (!ni) {
                notice(s_OperServ, source, "Nick %s isn't registered!", nick);
                return;
            }
		ni->flags &= ~NI_SOP;
                slog("!! Deleted \2%s\2 from SOP List", nick);
                do_break_log("OS", "!! Deleted %s from SOP List", nick);
		wallops(SERVER_NAME, "%s deleted %s from the SOP List", source, nick);
                notice(s_OperServ, source,
                        "%s removed from Services Operator list.", nick);
	}

    } else if (stricmp(cmd, "LIST") == 0) {
        NickInfo *ni;
        int i;
        int totalaccess = 0;
        notice(s_OperServ,source,"\2Services Operators:\2");
        for (i = 0; i < 256; ++i) {
            for (ni = nicklists[i]; ni; ni = ni->next) {
                if (ni->flags & NI_SOP) {
                  notice(s_OperServ,source,"%-20s",
                        ni->nick);
                  ++totalaccess;
                }
            }
        }
        notice(s_OperServ,source, NS_NFEOL, totalaccess);
    } else {
        notice(s_OperServ, source,
                "Syntax: \2SOP\2 [ADD|DEL|LIST] [<nick>]");
        notice(s_OperServ, source,
                "For help: \2/msg %s HELP SOP\2", s_OperServ);
    }
}
/***********************************************************************/
void do_set_oshop (const char *source) {
    char *cmd, *nick;
    NickInfo *ni;
    int i;

    cmd = strtok(NULL, " ");
    if (!cmd)
        cmd = "";
    if (stricmp(cmd, "ADD") == 0) {
        if (!is_services_admin(source)) {
            notice(s_OperServ, source, "Permission denied.");
            return;
        }
        nick = strtok(NULL, " ");
        if (nick) {
	    ni = findnick(nick);
            if (!ni) {
                notice(s_OperServ, source, "Nick %s isn't registered!", nick);
                return;
            }
	        ni->flags |= NI_HOP;
		slog("!! Added \2%s\2 to HOP List", nick);
		do_break_log("OS", "!! Added %s to HOP List", nick);
		wallops(SERVER_NAME, "%s added %s to the HOP List", source, nick);
		notice(s_OperServ, source, "%s added to Services Helper list.", nick);
	}
    } else if (stricmp(cmd, "DEL") == 0) {
        if (!is_services_admin(source)) {
            notice(s_OperServ, source, "Permission denied.");
            return;
        }
        nick = strtok(NULL, " ");
        if (nick) {
	    ni = findnick(nick);
            if (!ni) {
                notice(s_OperServ, source, "Nick %s isn't registered!", nick);
                return;
            }
		ni->flags &= ~NI_HOP;
                slog("!! Deleted \2%s\2 from HOP List", nick);
                do_break_log("OS", "!! Deleted %s from HOP List", nick);
		wallops(SERVER_NAME, "%s deleted %s from the HOP List", source, nick);
                notice(s_OperServ, source,
                        "%s removed from Services Helper list.", nick);
	}

    } else if (stricmp(cmd, "LIST") == 0) {
        NickInfo *ni;
        int i;
        int totalaccess = 0;
        notice(s_OperServ,source,"\2Services Helper:\2");
        for (i = 0; i < 256; ++i) {
            for (ni = nicklists[i]; ni; ni = ni->next) {
                if (ni->flags & NI_HOP) {
                  notice(s_OperServ,source,"%-20s",
                        ni->nick);
                  ++totalaccess;
                }
            }
        }
        notice(s_OperServ,source, NS_NFEOL, totalaccess);
    } else {
        notice(s_OperServ, source,
                "Syntax: \2SOP\2 [ADD|DEL|LIST] [<nick>]");
        notice(s_OperServ, source,
                "For help: \2/msg %s HELP HOP\2", s_OperServ);
    }
}
/***********************************************************************/
static void do_set_uin (NickInfo * ni, char *param)
{
       NickInfo *hni = host(ni);

       if (stricmp (param, "NONE") == 0) {
          hni->uin = NULL;
          notice (s_NickServ, ni->nick, "Your ICQ UIN was removed.");
       } else if ((strspn(param, "1234567890") == strlen(param)) && (strlen(param) > 5) && (strlen(param) < 11)) {
          hni->uin = sstrdup (param);
          notice (s_NickServ, ni->nick, "Change your ICQ UIN to : \2%s\2", param);
       } else {
          notice(s_NickServ, ni->nick, "Syntax: \2SET UNI\2 [\37uin\37|\37NONE\37]");
          notice(s_NickServ, ni->nick, "\2/msg %s HELP SET UIN\2 for more information.", s_NickServ);
       }
}

/**********************************************************************/
static void do_set_age (NickInfo * ni, char *param)
{
	NickInfo *hni = host(ni);
	if (!hni)
		return;

        if (stricmp (param, "NONE") == 0) {
          hni->age = NULL;
          notice (s_NickServ, ni->nick, "Your Age was removed.");
        } else if ((strlen(param) < 3) && (strspn(param, "1234567890") == strlen(param)))  {
          hni->age = sstrdup (param);
          notice (s_NickServ, ni->nick, "Change your AGE to : \2%s\2",param);
        } else {
          notice(s_NickServ, ni->nick, "Syntax: \2SET AGE\2 [\37age\37|\37NONE\37]");
          notice(s_NickServ, ni->nick, "\2/msg %s HELP SET AGE\2 for more information.", s_NickServ);
        }
}

/**********************************************************************/
static void do_set_info (NickInfo * ni, char *param)
{
    NickInfo *hni = host(ni);

    if (!hni)
	return;
    if (stricmp (param, "NONE") == 0)
    {
        hni->info = NULL;
        notice (s_NickServ, ni->nick, "Set your info to NULL.");
    }
    else
    {
	if (strlen(param) > 225) {
        notice (s_NickServ, ni->nick, "Your info can not set more than 225 characters.");
	} else {
        hni->info = sstrdup (param);
        notice (s_NickServ, ni->nick, "Change your info to :%s", param);
	}
    }
}
/**********************************************************************/
static void do_set_sex(NickInfo * ni, char *param)
{
      NickInfo *hni = host(ni);

      if (!hni)
	return;

      if (stricmp(param, "M") == 0)
      {
         hni->sex = "Male";
         notice (s_NickServ, ni->nick, "Change your GENDER to : \2Male\2");
      } else if (stricmp(param, "F") == 0) {
         hni->sex = "Female";
         notice (s_NickServ, ni->nick, "Change your GENDER to : \2Female\2");
      } else if (stricmp(param, "NONE") == 0) {
         hni->sex = NULL;
         notice (s_NickServ, ni->nick, "Your Gender was removed.");
      } else {
          notice (s_NickServ, ni->nick, "Syntax: \2SET SEX\2 [\37M\37|\37F\37|\37NONE\37]");
          notice (s_NickServ, ni->nick,  "\2/msg %s HELP SET SEX\2 for more information.", s_NickServ);
      }
}
/**********************************************************************/
static void do_set_ircop(NickInfo *ni, char *param)
{
   char *source = ni->nick;

   if (stricmp(param, "ON") == 0) {
       notice(s_NickServ, source, "Your IRCop flag turn on, your nick will never expired.");
       ni->flags |= NI_HOP;
       show_next_db(source, s_NickServ);
       return;

   } else if (stricmp(param, "OFF") == 0) {
       notice(s_NickServ, source, "Your IRCop flag turn off.");
       ni->flags &= ~NI_HOP;
       show_next_db(source, s_NickServ);
       return;

   } else {

       notice(s_NickServ, source, "Syntax: \2SET IRCOP\2 [\37ON\37|\37OFF\37]");
       notice(s_NickServ, source,
              "\2/msg %s HELP SET IRCOP\2 for more information.",
              s_NickServ);

  }
}
/**********************************************************************/
void do_listadm(const char *source)
{
    NickInfo *ni;
    int i;

    notice(s_OperServ, source, "\2Services Operators:\2");
    for (i = 0; i < 256; ++i) {
       for (ni = nicklists[i]; ni; ni = ni->next)
           if (ni->flags & NI_SOP)
               notice(s_OperServ, source, "%s (%s)", ni->nick, ni->last_usermask);
    }

    notice(s_OperServ, source, "\2Services Administrators:\2");
    for (i = 0; i < MAX_SERVADMINS; i++) {
            if (services_admins[i]) {
              if (ni = findnick(services_admins[i]))
                 notice(s_OperServ, source, "%s (%s)", services_admins[i], ni->last_usermask);
            }
    }
    notice(s_OperServ, source, "\2Services Root Administrators:\2");
    for (i = 0; i < MAX_SERVROOTS; i++) {
        if (services_roots[i]) {
            if (stricmp(services_roots[i], SERVICES_MASTER) == 0){
                   if (ni = findnick(services_roots[i]))
                      notice(s_OperServ, source, "%s (%s) [Master]", services_roots[i], ni->last_usermask);
            } else if (stricmp(services_roots[i], SERVICES_MASTER2) == 0){
                   if (ni = findnick(services_roots[i]))
                      notice(s_OperServ, source, "%s (%s) [Co Master]", services_roots[i], ni->last_usermask);
            } else if (stricmp(services_roots[i], SERVICES_MASTER3) == 0){
                   if (ni = findnick(services_roots[i]))
                      notice(s_OperServ, source, "%s (%s) [Co Master]", services_roots[i], ni->last_usermask);
            } else {
              if (ni = findnick(services_roots[i]))
                 notice(s_OperServ, source, "%s (%s)", services_roots[i], ni->last_usermask);
            }
        }
    }

    notice(s_OperServ, source, ERR_EOL);
}

/**************************************************************************/

/**************************************************************************/
/*
static void do_set_mlock(NickInfo * ni, char *param)
{
    char *s, c, *nmlock_on[MODEMAX], *nmlock_off[MODEMAX], *buf, *modes;
    int add = -1;

    if (!param)
        notice (s_NickServ, ni->nick, "Syntax: \2SET MLOCK umodes\2");
    else if ((ni->flags & NI_MARK) && (!is_services_admin(ni->nick)))
        notice (s_NickServ, ni->nick, "Can not SET MLOCK on marked nick. ");
    else
    {
	strscpy(nmlock_on, "", sizeof(nmlock_on));
	strscpy(nmlock_off, "", sizeof(nmlock_off));
    while (*param)
    {
	slog("not even one");
        if (*param != '+' && *param != '-' && add < 0)
        {
            ++param;
            continue;
        }
        switch ((c = *param++))
        {
        case '+':
            add = 1;
            break;
        case '-':
            add = 0;
            break;
        case 'r':
            break;
	case 'G' || 's' || 'i' || 'B' || 'w':
            if (add)
		strscpy(nmlock_on, changemode(myctoa(c), nmlock_on), sizeof(nmlock_on));
            else
		strscpy(nmlock_off, changemode(myctoa(c), nmlock_off), sizeof(nmlock_off));
	    break;
	default:
	    break;
	}
    }
       sprintf (modes,"+%s", nmlock_on);
       strcat(buf, modes);
       *modes = 0;
       sprintf (modes,"-%s", nmlock_off);
       strcat(buf, modes);
       ni->mlock = sstrdup(buf);
       show_next_db(ni->nick, s_NickServ);
       notice(s_NickServ, ni->nick, "Mlock changed to \2%s\2.", ni->mlock);
    }
}
*/
